/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#include "analyser.h"
#include "xxx.h"
#include <cassert>
#include <cstdio>
#include <map>
#include <set>

/* This module analyses the XML input tree and creates the
 * ProblemSpecification object. To use it you have to create AnalyserClass
 * object (which is subclass of ProblemSpecification) and pass the XML tree
 * as argument of constructor. 
 * 
 * The another functionality is to analyse XML tree with solution, to do it
 * call method AddSolution (of cource problem specification is needed to
 * analyse the solution). It saves the solution in 'solution'. You can call
 * the method several times. */

using namespace XML;

namespace Analyser
{

void AnalyserClass::GetXMLAtts(const vector<Attribute> &attributes, const AttDef *defs, int defsNo, int line)
/* Analyses attributes in 'attributes' using definition of possible
 * attributes in 'defs'. 'defsNo' is number of elements in 'defs'. */
{
	vector<int> occur(defsNo, 0);
	FOREACH(a, attributes)
	{
		int b;
		for (b = 0; b<defsNo; ++b)
			if (strcmp(defs[b].name, a.name.c_str())==0)
				goto found;
		yyerror_arg(a.line, "Unexpected attribute name: \"%s\"", a.name.c_str());
	found:
		++occur[b];
		if (occur[b]>1)
			yyerror_arg(a.line, "Repeated attribute declaration \"%s\"", a.name.c_str());
		if (defs[b].function)
			(this->*defs[b].function)(a);
	}
	REP(b, defsNo)
		if (defs[b].type==AttDef::one && !occur[b])
			yyerror_arg(line, "Missing attribute declaration: \"%s\"", defs[b].name);
}

void AnalyserClass::NoAtts(const vector<Attribute> &attributes)
/* Call it inside element analysing function to check that the element do
 * not have attributes. */
{
	if (!attributes.empty())
		yyerror_arg(attributes[0].line, "Unexpected attribute name: \"%s\"", attributes[0].name.c_str());
}

void AnalyserClass::GetXMLSons(const Content *inDesc0, const SonDef *defs, int defsNo, int line)
/* Analyses subelements in 'inDesc0' using definition of possible sons in
 * 'defs'. 'defsNo' is number of elements in 'defs'. */
{
	vector<int> occur(defsNo, 0);
	if (inDesc0!=NULL)
	{
		CREF(inDesc, inDesc0->GetElements());
		FOREACH(a, inDesc.children)
		{
			int b;
			for (b = 0; b<defsNo; ++b)
				if (strcmp(defs[b].name, a.name.c_str())==0)
					goto found;
			yyerror_arg(a.line, "Unexpected name: \"%s\"", a.name.c_str());
		found:
			occur[b]++;
			if (defs[b].type!=SonDef::anyNumber && occur[b]>1)
				yyerror_arg(a.line, "Repeated declaration \"%s\"", a.name.c_str());
			if (defs[b].function)
				(this->*defs[b].function)(a.content, a.attributes, a.line);
		}
	}
	REP(b, defsNo)
		if (defs[b].type==SonDef::one && !occur[b])
			yyerror_arg(inDesc0!=NULL?inDesc0->GetElements().endline:line, 
				"Missing declaration: \"%s\"", defs[b].name);
}

void AnalyserClass::analyseResource(const Content *inDesc, const vector<Attribute> &attributes, int line)
{
	int s = size(resources);
	resources.resize(s+1);
	REF(r, resources[s]);
	analyserInfo->id = &r.id;
	analyserInfo->timePreferences = &r.timePreferences;
	analyserInfo->resource = &r;
	r.breakLen = 0;
	r.maxDays = INF;
	r.balancedScore = 0;
	CONT_ONLY( r.continuousScore = 0; )
	analyseId(inDesc, attributes, line);
	analyseResource2(inDesc, attributes, line);
} 

void AnalyserClass::analyseGroup(const Content *inDesc, const vector<Attribute> &attributes, int line)
{
	analyserInfo->conjunctive = 0;
	analyseGroupType(inDesc, attributes, line);
	if (!analyserInfo->conjunctive)
	{
		REF(v, analyserInfo->activity->temporaryData->requiredGroups);
		int s = size(v);
		v.resize(s+1);
		analyserInfo->resourcesList = &v[s];
	}
	/* Resources from disjunctive group are simply added to the list of
	 * required resources. */
	analyseGroup2(inDesc, attributes, line);
	if (!analyserInfo->conjunctive)
	{
		REF(v, analyserInfo->activity->temporaryData->requiredResources);
		CREF(v2, *analyserInfo->resourcesList);
		if (v2.empty()) yyerror_arg(line, "Empty group");
		if (size(v2)==1)
		/* Group with one resource is not a group. */
		{
			v.push_back(v2[0]);
			vector<vector<string> > &v3 = analyserInfo->activity->temporaryData->requiredGroups;
			v3.resize(size(v3)-1);
		}
		analyserInfo->resourcesList = &v;
	}
}

void AnalyserClass::analyseActivity(const Content *inDesc, const vector<Attribute> &attributes, int line)
{
	int s = size(activities);
	activities.resize(s+1);
	REF(a, activities[s]);
	a.Initialize();
	WEEKS_ONLY( a.weekFlag = ALL_WEEKS; )
	analyserInfo->id = &a.id;
	analyserInfo->timePreferences = &a.timePreferences;
	analyserInfo->activity = &a;
	analyserInfo->resourcesList = &a.temporaryData->requiredResources;
	analyseId(inDesc, attributes, line);
	analyseActivity2(inDesc, attributes, line);
}

void AnalyserClass::analyseDependence(const Content *inDesc, const vector<Attribute> &attributes, int line)
{
	int s = size(dependences);
	dependences.resize(s+1);
	REF(d, dependences[s]);
	analyserInfo->dependence = &d;
	d.score = -1;
	analyserInfo->score = &d.score;
	string skip;
	analyserInfo->id = &skip;
	analyseIdScore(inDesc, attributes, line);
	analyseDependence2(inDesc, attributes, line);
}

void AnalyserClass::analyseActivity3(const Content *inDesc, const vector<Attribute> &attributes, int line)
{
	int s = size(solution);
	solution.resize(s+1);
	REF(a, solution[s]);
	analyserInfo->id = &a.id;
	analyserInfo->resourcesList = &a.usedResourcesStr;
	analyserInfo->startTime = &a.startTime;
	analyseId(inDesc, attributes, line);
	if (inDesc==NULL)
		a.startTime = -1;
	else
		analyseActivity4(inDesc, attributes, line);
}

void AnalyserClass::ReadInt(const string &value, int &var, int line)
/* Read int from 'value' and save it to 'var' (or show error). */
{
	FOREACH(a, value)
		if (a>'9' || a<'0')
			goto fail;
	if (sscanf(value.c_str(), "%d", &var)==1) 
		return;
fail:
	yyerror_arg(line, "Integer number required (got \"%s\")", value.c_str());
}

void AnalyserClass::ReadInterval(const string &value, vector<int> &var, int line)
/* Read interval in form "number#number" or "number" (or show error). All
 * values from it will be added to 'var' vector. */
{
	int a;
	int x, y;
	for (a = 0; a<size(value); ++a)
		if (value[a]=='#')
			goto ok;
	ReadInt(value, x, line);
	var.push_back(x);
	return;
ok:
	ReadInt(value.substr(0, a), x, line);
	ReadInt(value.substr(a+1), y, line);
	if (x>y)
		yyerror_arg(line, "Interval reversed (\"%s\")", value.c_str());
	REP(c, y-x+1)
		var.push_back(x+c);
}


void AnalyserClass::TimePreferences::Normalize(int oldNoOfSlots, int oldSlotsPerDay, int addSlots)
/* Normalyse and check correctness of TimePreferences structure. Add
 * 'addSlots' forbiden slots in each day. Declared in problemspec.h. See
 * there for more info. */
{
	int daysPerWeek = oldNoOfSlots/oldSlotsPerDay;
	int newSlotsPerDay = oldSlotsPerDay+addSlots;
	int newNoOfSlots = newSlotsPerDay*daysPerWeek;
	#define old2new(a) ((a)/oldSlotsPerDay*newSlotsPerDay+(a)%oldSlotsPerDay)
	vector<int> v;
	v.resize(newNoOfSlots, hardMeansUnavail);
	FOREACH(ai, hard)
	{
		if (ai>=oldNoOfSlots)
			yyerror_arg("Timeslot number out of range (%d)", ai);
		v[old2new(ai)] = !hardMeansUnavail;
	}
	REP(d, daysPerWeek)
		REP(a, addSlots)
			v[d*newSlotsPerDay+oldSlotsPerDay+a] = 0;
	hard.clear();
	REP(a, newNoOfSlots)
		if (!v[a])
			hard.push_back(a);
	hardMeansUnavail = 1;
	REP(a, newNoOfSlots)
		v[a] = 0;
	for (int a = 0, s = 0; a<size(soft); ++a)
		if (soft[a]<0)
			s = -soft[a];
		else
		{
			if (soft[a]>=oldNoOfSlots)
				yyerror_arg("Timeslot number out of range (%d)", soft[a]);
			v[old2new(soft[a])] += s;
		}
	soft = v;
	#undef old2new
}

void AnalyserClass::TimePreferences::CompressSoft(int length)
/* Calculates 'compressedSoft': compressedSoft[t] = sum of soft from t to
 * t+length-1. */
{
	compressedSoft.resize(size(soft));
	compressedSoft[0] = 0;
	assert(length<=size(soft));
	REP(a, length)
		compressedSoft[0] += soft[a];
	FOR(t, 1, size(soft)-length)
		compressedSoft[t] = compressedSoft[t-1]+soft[t+length-1]-soft[t-1];
}

void AnalyserClass::NormalizeTimePreferences()
/* Normalyse and check correctness of time preferences for all activities
 * and resources (for more info see problemspec.h). Set
 * 'additionalTimeslots' to maximum length of break and correct time
 * preferences. Additionaly copy 'timePreferences' to
 * 'initialTimePreferences' and calculate 'mdLength'. */
{
	FOREACH(r, resources)
		updateMax(additionalTimeslots, r.breakLen);
	if (additionalTimeslots)
		show(10,("Added additional %d timeslots at end of each day\n", additionalTimeslots));
	FOREACH(r, resources)
		r.timePreferences.Normalize(noOfSlots, slotsPerDay, additionalTimeslots);
	FOREACH(a, activities)
	{
		a.timePreferences.Normalize(noOfSlots, slotsPerDay, additionalTimeslots);
		a.initialTimePreferences = a.timePreferences;
		a.mdLength = a.length*WEEKS_ONLY2((a.weekFlag==ALL_WEEKS?2:1), 2);
	}
	noOfSlots += additionalTimeslots*daysPerWeek;
	slotsPerDay += additionalTimeslots;
}

int AnalyserClass::GetNoById(const map<string,int> &m, const string &str)
/* Return m[str] or print error when not found. */
{
	VAR(res, m.find(str));
	if (res==m.end()) 
		yyerror_arg("Id \"%s\" not found", str.c_str());
	return res->second;
}

int AnalyserClass::Group::operator<(const AnalyserClass::Group& gr2) const
/* Smaller groups are less than biger. Groups of the same size are in
 * lexicographical order of resources.  Comparation is used for eliminating
 * identical groups. Declared in problemspec.h. */
{
	int l1 = size(resources);
	int l2 = size(gr2.resources);
	if (l1<l2) return 1;
	if (l2<l1) return 0;
	for (vector<int>::const_iterator a = resources.begin(), b = gr2.resources.begin(); 
	     a<resources.end(); ++a, ++b)
		if (*a<*b) return 1; 
		else 
		if (*b<*a) return 0;
	return 0;
}

AnalyserClass::TimePreferences &AnalyserClass::TimePreferences::operator+=
    (const AnalyserClass::TimePreferences &tp2)
/* Sum of TimePreferences means that both hard preferences must hold and
 * score for soft preferences is sum of score from each. The operator works
 * correctly in normalized form. Declared in problemspec.h. */
{
	vector<int> hard1 = hard;
	CREF(soft2, tp2.soft);
	CREF(hard2, tp2.hard);
	hard.resize(0);
	vector<int>::iterator i1;
	vector<int>::const_iterator i2;
	for (i1 = hard1.begin(), i2 = hard2.begin(); i1<hard1.end() || i2<hard2.end(); )
	{
		if (i1==hard1.end()) { hard.push_back(*(i2++)); continue; }
		if (i2==hard2.end()) { hard.push_back(*(i1++)); continue; }
		if (*i1==*i2) { ++i1; continue; }
		if (*i1<*i2) hard.push_back(*(i1++));
		else hard.push_back(*(i2++));
	}
	for (i1 = soft.begin(), i2 = soft2.begin(); i1<soft.end(); ++i1, ++i2)
		*i1 += *i2;
	return *this;
}

template<class R> void createIdMap(map<string, int> &m, const vector<R> &vec)
/* 'vec' should be vector of 'Resource's or 'Activity's. Creates map from
 * resources or activities IDs to its positions in vector. */
{
	REPV(r, vec)
	{
		if (m.find(vec[r].id)!=m.end())
			yyerror_arg("The same ID used twice (%s)", vec[r].id.c_str());
		m[vec[r].id] = r;
	}
}

void AnalyserClass::GenerateRequiresAndGroups()
/* Generate 'requiredResources', 'inActivitiesRe', 'groups',
 * 'requiredGroups', 'inGroups', 'inActivitesGr',
 * 'activities.timePreferences' and deletes 'temporaryData'. Earlier time
 * preferences should be normalized. */
{
	show(40, ("  Generating requires and groups\n"));
	map<string,int> m;	// Map from resources IDs to numbers.
	map<Group,int> gr_set;	// Map from groups to its numbers.
	int grno = 0;	// Current numbers of groups. Increases when new
			// groups are added.
	createIdMap(m, activities);	// Just check if no ID twice.
	m.clear();
	createIdMap(m, resources);
	REPV(aind, activities)
	{
		show(40, ("    Activity %d [%s]\n", ACT(aind)));
		REF(a, activities[aind]);
		set<int> usedRes;	// Resources used by current
					// activity (in groups or not).
		show(40, ("      Required resources:"));
		FOREACH(r, a.temporaryData->requiredResources)
		{
			int rnr = GetNoById(m, r);
			show(40, (" %d [%s]", RES(rnr)));
			a.requiredResources.push_back(rnr);
			resources[rnr].inActivitiesRe.push_back(aind);
			WEEKS_ONLY(
			if (a.weekFlag!=ALL_WEEKS)
			{
				TimePreferences tp = resources[rnr].timePreferences;
				FOREACH(sp, tp.soft) sp/=2;
				a.timePreferences += tp;
			}
			else 
			)
				a.timePreferences += resources[rnr].timePreferences;
			if (usedRes.find(rnr)!=usedRes.end())
				yyerror_arg("The same resource needed twice (%d [%s])", RES(rnr));
			usedRes.insert(rnr); 
		}
		show(40, ("\n"));
		a.timePreferences.CompressSoft(a.length);
		show(40, ("      Required groups:"));
		FOREACH(g, a.temporaryData->requiredGroups)
		{
			Group gr;
			show(40, (" ("));
			FOREACH(r, g)
			{
				show(40, (" %d [%s]", RES(GetNoById(m, r))));
				gr.resources.push_back(GetNoById(m, r));
			}
			VAR(found, gr_set.find(gr));
			if (found==gr_set.end())	// new Group
			{
				show(40, (" :new:"));
				FOREACH(r, gr.resources)
					resources[r].inGroups.push_back(grno);
				found = gr_set.insert(make_pair(gr, grno++)).first;
			}
			show(40, (")"));

			/* Here we cast to non const, but it is allowed
			 * because we do not change the order of Groups in
			 * the map. */
			const_cast<AnalyserClass::Group&>(found->first).activities.push_back(aind);

			a.requiredGroups.push_back(found->second);
			
			FOREACH(r, gr.resources)
			{
				if (usedRes.find(r)!=usedRes.end())
					yyerror_arg("The same resource needed twice (%d [%s])", RES(r));
				usedRes.insert(r);
			}
		}
		show(40, ("\n"));
		delete a.temporaryData;
	}
	groups.resize(size(gr_set));
	FOREACH(g, gr_set)
		groups[g.second] = g.first;
}

void AnalyserClass::NormalizeDependences()
/* Normalize dependences (reverse after and closelyAfter dependences and
 * cocurrenly dependences with first activity shorter than second) and
 * change activities IDs to numbers (in dependences). Calculate
 * 'inDependences' for activities. */
{
	if (dependences.empty()) return;
	map<string,int> m;
	createIdMap(m, activities);
	REPV(di, dependences)
	{
		REF(d, dependences[di]);
		d.firstAct = GetNoById(m, d.firstActStr);
		d.firstActStr.resize(0);
		d.secondAct = GetNoById(m, d.secondActStr);
		d.secondActStr.resize(0);
		if (d.relation==Dependence::after || d.relation==Dependence::closelyAfter || 
		    (d.relation==Dependence::concurrently && activities[d.firstAct].length<activities[d.secondAct].length))
		{
			if (d.relation==Dependence::after) d.relation = Dependence::before;
			if (d.relation==Dependence::closelyAfter) d.relation = Dependence::closelyBefore;
			swap(d.firstAct, d.secondAct);
		}
		activities[d.firstAct].inDependences.push_back(di);
		activities[d.secondAct].inDependences.push_back(di);
	}
}

void AnalyserClass::GenerateGrres()
/* Calculate 'grresNo', 'contresNo', 'mdresNo', 'grres', 'contres', 'mdres',
 * 'contCount', 'averageUsage' for all resources. Earlier 'inGroups' and
 * 'inActivietiesRe' should be calculated. */
{
	mdresNo = grresNo = CONT_ONLY(contresNo =) 0;
	REPV(rn, resources)
	{
		REF(r, resources[rn]);
		r.grres = r.inGroups.empty() ? -1 : grresNo++;
		r.mdres = r.maxDays==INF&&r.balancedScore==0 ? -1 : mdresNo++;
		if (r.maxDays<1)
			yyerror_arg("MaxDays shoud be at least 1 (resource %d [%s])", RES(rn));
		if (r.mdres>=0 && r.grres>=0)
			yyerror_arg("MaxDays and Balanced are not supported for resources in groups (resource %d [%s])", RES(rn));
		r.averageUsage = 0;
		FOREACH(a, r.inActivitiesRe)
			r.averageUsage += activities[a].length*
				WEEKS_ONLY2((activities[a].weekFlag==ALL_WEEKS?2:1), 2);
		r.averageUsage /= min(daysPerWeek, r.maxDays);
		CONT_ONLY(
		r.contres = r.continuousScore==0 ? -1 : contresNo++;
		if (r.contres>=0 && r.grres>=0)
			yyerror_arg("Continuous not supported for resources in groups (resource %d [%s])", RES(rn));
		if (r.contres>=0)
		{
			r.contCount WEEKS_ONLY([0] = r.contCount[1]) = 0;
			FOREACH(a, r.inActivitiesRe)
			{
				CREF(act, activities[a]);
				WEEKS_ONLY2(
				  if (act.weekFlag&EVEN_WEEKS) r.contCount[0] += act.length+r.breakLen;
				  if (act.weekFlag&ODD_WEEKS) r.contCount[1] += act.length+r.breakLen;
				,
				  r.contCount += act.length;
				)
			}
		}
		)
	}
}

void AnalyserClass::GenerateGrsel()
/* Calculate grselNo and grsel for all activities. Earlier 'requiredGroups'
 * should be calculated. */
{
	grselNo = 0;
	FOREACH(a, activities)
	{
		a.grsel = grselNo;
		grselNo += size(a.requiredGroups);
	}
}

void AnalyserClass::GenerateConflicts()
/* Calculates 'conflictsWith' in activities. Earlier 'requiredResources' and
 * 'inActivitiesRe' should be calculated. */ 
{
	FOREACH(a, activities)
	{
		map<int,int> s;
		FOREACH(r, a.requiredResources)
			FOREACH(a2, resources[r].inActivitiesRe)
			{
				WEEKS_ONLY( if (a.weekFlag&activities[a2].weekFlag) )
					updateMax(s[a2], resources[r].breakLen);
			}
		FOREACH(a2, s)
			a.conflictsWith.push_back(a2);
	}
}

void AnalyserClass::Initialize(const Content *inDesc)
/* Takes the XML tree and does all the analysis of input. */
{
	additionalTimeslots = 0;
	vector<Attribute> emptyvec;
	analyserInfo = new AnalyserInfo();
	analyse(inDesc, emptyvec, 1);
	delete analyserInfo;
	noOfSlots = slotsPerDay*daysPerWeek;
	NormalizeTimePreferences();
	GenerateRequiresAndGroups();
	NormalizeDependences();
	GenerateGrres();
	GenerateGrsel();
	GenerateConflicts();
	show(10,("Statistics: %d activities, %d resources, %d groups, %d timeslots (%d days x %d hours), %d dependences, %d grres, %d grsel\n",
		size(activities), size(resources), size(groups), noOfSlots, daysPerWeek, slotsPerDay, size(dependences), grresNo, grselNo));
}

void AnalyserClass::FindStringsInSolution()
/* This method does all the postprocessing while analysing solution. It
 * converts IDs of activities and resources to its numbers. It also checks
 * if we have every activity exactly once. */
{
	map<string,int> mr, ma;
	createIdMap(ma, activities);
	createIdMap(mr, resources);
	vector<bool> act(size(activities), 0);
	FOREACH(sa, solution)
	{
		sa.actNo = GetNoById(ma, sa.id);
		if (act[sa.actNo])
			yyerror_arg("Two times the same activity %d [%s] (in solution)", ACT(sa.actNo));
		sa.id.clear();
		act[sa.actNo] = true;
		FOREACH(ri, sa.usedResourcesStr)
			sa.usedResources.push_back(GetNoById(mr, ri));
		sa.usedResourcesStr.clear();
	}
	if (size(solution)!=size(activities))
		REPV(a, act)
			if (!act[a])
				yyerror_arg("Missing activity %d [%s] (in solution)", ACT(a));
}

void AnalyserClass::AdditionalTimeslotsInSolution()
/* Convert times in 'solution' from normal to these with
 * 'additionalTimeslots' at end of each day. */
{
	FOREACH(sa, solution)
		sa.startTime += sa.startTime/(slotsPerDay-additionalTimeslots)*additionalTimeslots;
}

void AnalyserClass::AddSolution(const Content *inDesc)
/* This method analyses the solution XML tree. It saves it in 'solution'. 
 * You can call the method several times. */
{
	vector<Attribute> emptyvec;
	solution.resize(0);
	analyserInfo = new AnalyserInfo();
	analyse2(inDesc, emptyvec, 1);
	delete analyserInfo;
	FindStringsInSolution();
	AdditionalTimeslotsInSolution();
}

} // namespace Analyser

