/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#ifndef __ANALYSER_H
#define __ANALYSER_H

#include "inputparser_my.h"
#include "xml.h"
#include "tools.h"
#include "problemspec.h"
#include <vector>
#include <map>

/* This module analyses the XML input tree and creates the
 * ProblemSpecification object. To use it you have to create AnalyserClass
 * object (which is subclass of ProblemSpecification) and pass the XML tree
 * as argument of constructor. 
 * 
 * The another functionality is to analyse XML tree with solution, to do it
 * call method AddSolution (of cource problem specification is needed to
 * analyse the solution). It saves the solution in 'solution'. You can call
 * the method several times. */

using namespace std;

namespace Analyser
{

class AnalyserClass : public ProblemSpecification
{
private:
	struct AnalyserInfo
	{
		/* This object contains pointers to some data which will be
		 * read. For example 'id' can be pointer to 'id' of current
		 * activity or resource. Procedure reading the ID uses this
		 * pointer and do not have to know whether we are in
		 * resource or in activity and in which one. This object
		 * exists only while analysing. */

		string *id;
		TimePreferences *timePreferences;
		Resource *resource;
		Activity *activity;
		Dependence *dependence;
		int *score;
		int *startTime;
		vector<string> *resourcesList;
		int conjunctive;
	};
	AnalyserInfo *analyserInfo;

	/* Here is a lot of macros which helps in writing the analyser. */
	
	// Analysing subelements /////////////////////////////////////////

	#define AnalFuncArgs (const XML::Content *inDesc, const vector<XML::Attribute> &attributes, int line)
		// Arguments of typical function analysing element.
	
	typedef void (AnalyserClass::*AnalyseFuncType) AnalFuncArgs;
		// Type of typical function analysing element.
		
	struct SonDef
	{
		/* Definition of posible sons of the element. For each
		 * 'name' we give 'function' which will be called to analyse
		 * this subelement. We also define the 'type'. */
		const char *name;
		AnalyseFuncType function;
		enum Type { zeroOrOne, one, anyNumber } type;
		SonDef(const char *str, AnalyseFuncType f, enum Type t = one) : name(str), function(f), type(t) {}
	};

	#define GetSons() GetXMLSons(inDesc, sonDef, sizeof(sonDef)/sizeof(sonDef[0]), line)
	/* Call at end of function, after possible son elements are defined
	 * in 'sonDef'. It will analyse the sons. */

	/* 'Anal...' macros starts the analysing functions. After 'AnalFunc'
	 * you can mannualy write the function. After the other two you
	 * should use 'Son...' macros to define sons, these function should
	 * end with 'EndAnal'. 'Anal' checks if the element does not have
	 * attributes. In 'AnalAA' element can have attributes, but they are
	 * not analysed here. */

	#define AnalFunc(str) void analyse##str AnalFuncArgs
	#define Anal(str) AnalFunc(str) { NoAttr; SonDef sonDef[] = { 
	#define AnalAA(str) AnalFunc(str) { SonDef sonDef[] = { 
	#define EndAnal }; GetSons(); }

	/* Use the 'Son...' macros between 'Anal' or 'AnalAA' and 'EndAnal'
	 * to define possible subelements. Macros with '2' allows to give
	 * other function name than element name (use when element with the
	 * same can appear in different places and has different meanings). 
	 * 'Son' and 'Son2' says that the son has to appear exactly once. 
	 * 'Son?Opt' allows to appear at most once. 'Son?Any' allows the son
	 * to appear any number of times. The 'SonOptSkip' defines a son
	 * which can appear or not, but it will not be analysed. */

	#define Son2(name,fname) SonDef(#name, &AnalyserClass::analyse##fname)
	#define Son2Opt(name,fname) SonDef(#name, &AnalyserClass::analyse##fname, SonDef::zeroOrOne)
	#define Son2Any(name,fname) SonDef(#name, &AnalyserClass::analyse##fname, SonDef::anyNumber)
	#define Son(name) Son2(name,name)
	#define SonOpt(name) Son2Opt(name,name)
	#define SonAny(name) Son2Any(name,name)
	#define SonOptSkip(name) SonDef(#name, NULL, SonDef::zeroOrOne)

	void GetXMLSons(const XML::Content *inDesc0, const SonDef *defs, int defsNo, int line);
	/* Analyses subelements in 'inDesc0' using definition of possible
	 * sons in 'defs'. 'defsNo' is number of elements in 'defs'. */

	// Analysing attributes ---------------------------------------------

	#define AnalAttFuncArgs (const XML::Attribute &attribute)
		// Arguments of typical function analysing attributes.

	typedef void (AnalyserClass::*AnalyseAttFuncType) AnalAttFuncArgs;
		// Type of typical function analysing attributes.

	struct AttDef
	{
		/* Definition of posible attributes of the element. For each
		 * 'name' we give 'function' which will be called to analyse
		 * this attribute. We also define the 'type'. */
		const char *name;
		AnalyseAttFuncType function;
		enum Type { zeroOrOne, one } type;
		AttDef(const char *str, AnalyseAttFuncType f, enum Type t = one) : name(str), function(f), type(t) {}
	};

	#define GetAtts() GetXMLAtts(attributes, attDef, sizeof(attDef)/sizeof(attDef[0]), line)
	/* Call at end of function, after possible attributes are defined
	 * in 'attDef'. It will analyse the attributes. */

	
	#define AnalAttFunc(str) void analyseAtt##str AnalAttFuncArgs
	/* 'AnalAttFunc' defines a function which analyses SINGLE ATTRIBUTE. 
	 *  After that you can mannualy write the function. */

	/* 'AnalAtt' macro starts the attributes analysing function.  After
	 * that you should use the 'Att...' macros to define attributes, the
	 * function should end with 'EndAnalAtt'. */

	#define AnalAtt(str) AnalFunc(str) { AttDef attDef[] = { 
	#define EndAnalAtt }; GetAtts(); }

	/* Use the 'Att...' macros between 'AnalAtt' and 'EndAnalAtt' to
	 * define possible attributes.  'Att' says that the attribute has to
	 * appear exactly once.  'AttOpt' allows it to appear at most once.
	 */

	#define Att(name) AttDef(#name, &AnalyserClass::analyseAtt##name)
	#define AttOpt(name) AttDef(#name, &AnalyserClass::analyseAtt##name, AttDef::zeroOrOne)

	void GetXMLAtts(const vector<XML::Attribute> &attributes, const AttDef *defs, int defsNo, int line);
	/* Analyses attributes in 'attributes' using definition of possible
	 * attributes in 'defs'. 'defsNo' is number of elements in 'defs'.
	 */

	#define NoAttr NoAtts(attributes)
	void NoAtts(const vector<XML::Attribute> &attributes);
	/* Call it inside element analysing function to check that the
	 * element do not have attributes. */

	// Analysing values -------------------------------------------------

	const string &XMLGetText(const XML::Content *inDesc, int line)
	/* Return the value of text content of a element (or error). */
	{
		if (inDesc==NULL)
			yyerror_arg(line, "Empty element");
		return inDesc->GetText().value;
	}
	
	#define AnalInt(what,var) AnalFunc(what) { NoAttr; ReadInt(XMLGetText(inDesc, line), var, line); }
	/* It define the analysing function for a element which should
	 * contain a integer. It will be read to 'var'. */
	
	void ReadInt(const string &value, int &var, int line);
	/* Read int from 'value' and save it to 'var' (or show error). */

	void ReadInterval(const string &value, vector<int> &var, int line);
	/* Read interval in form "number#number" or "number" (or show
	 * error). All values from it will be added to 'var' vector. */

	// ------------------------------------------------------------------

	AnalInt(DaysPerWeek, daysPerWeek)
	AnalInt(SlotsPerDay, slotsPerDay)

	Anal(General)
		Son(DaysPerWeek),
		Son(SlotsPerDay),
		SonOptSkip(NbTimeSlots),
		SonOptSkip(TimeSlots),
		SonOptSkip(Resources),
		SonOptSkip(Activities),
		SonOptSkip(Dependences)
	EndAnal

	AnalAttFunc(id)	// --> analyserInfo->id
	{
		*analyserInfo->id = attribute.value;
	}

	AnalAtt(Id)
		Att(id)
	EndAnalAtt

	AnalAttFunc(score)	// --> analyserInfo->score
	{
		ReadInt(attribute.value, *analyserInfo->score, attribute.line);
	}

	AnalAtt(Score)
		AttOpt(score)
	EndAnalAtt

	AnalAtt(IdScore)
		AttOpt(score),
		AttOpt(id)
	EndAnalAtt

	AnalFunc(Soft)
	{
		int s = DEFAULT_SOFT_SCORE;
		analyserInfo->score = &s;
		analyseScore(inDesc, attributes, line);
		if (!s) yyerror_arg(line, "Score is zero?");
		analyserInfo->timePreferences->soft.push_back(-s);
		ReadInterval(XMLGetText(inDesc, line), analyserInfo->timePreferences->soft, line);
	}

	AnalFunc(Hard)
	{
		NoAttr;
		TimePreferences &tp = *analyserInfo->timePreferences;
		if (tp.hardMeansUnavail==0)
			yyerror_arg(line, "'Hard' and 'Avail' can not apear simultanously");
		tp.hardMeansUnavail = 1;
		ReadInterval(XMLGetText(inDesc, line), tp.hard, line);
	}

	AnalFunc(Avail)
	{
		NoAttr;
		TimePreferences &tp = *analyserInfo->timePreferences;
		if (tp.hardMeansUnavail==1)
			yyerror_arg(line, "'Hard' and 'Avail' can not apear simultanously");
		tp.hardMeansUnavail = 0;
		ReadInterval(XMLGetText(inDesc, line), tp.hard, line);
	}

	Anal(TimePreferences)	// --> analyser->timePreferences
		SonAny(Hard),
		SonAny(Soft),
		SonAny(Avail)
	EndAnal

	CONT_ONLY(
	AnalFunc(Continuous)
	{
		analyserInfo->score = &analyserInfo->resource->continuousScore;
		*analyserInfo->score = CONTINUOUS_SCORE;
		analyseScore(inDesc, attributes, line);
	}
	)

	AnalFunc(Balanced)
	{
		analyserInfo->score = &analyserInfo->resource->balancedScore;
		analyseScore(inDesc, attributes, line);
		if (analyserInfo->score==0)
			yyerror_arg(line, "Balanced needs score attribute (!=0)");
	}
	
	AnalInt(Break, analyserInfo->resource->breakLen)

	AnalInt(MaxDays, analyserInfo->resource->maxDays)

	AnalAA(Resource2)	// Resource in Resources, but these function
				// is called inside analyseResource which
				// does some more additional computation.
				// --> analyserInfo->resource
		SonOpt(TimePreferences),
		#if ENABLE_CONT
		SonOpt(Continuous),
		#endif
		SonOpt(Break),
		SonOpt(MaxDays),
		SonOpt(Balanced),
		SonOptSkip(Name),
		SonOptSkip(Year),
		SonOptSkip(Type)
	EndAnal

	AnalFunc(Resource);	// Resource in Resources
				// --> analyserInfo->resourcesList

	Anal(Res)		// one of Classrooms, Teachers, Classes,
				// Special in Problem
				// --> analyserInfo->resourcesList
		SonAny(Resource)
	EndAnal

	Anal(Resources)	// Resources in Problem
		Son2(Classrooms,Res),
		Son2(Teachers,Res),
		Son2(Classes,Res),
		Son2Opt(Special,Res),
	EndAnal

	AnalInt(Length, analyserInfo->activity->length);

	WEEKS_ONLY(
	AnalFunc(Week)
	{
		NoAttr;
		const string &str = XMLGetText(inDesc, line);
		if (str.length()!=1) goto fail;
		switch(str[0])
		{
		case 'A': analyserInfo->activity->weekFlag = ALL_WEEKS; return;
		case 'E': analyserInfo->activity->weekFlag = EVEN_WEEKS; return;
		case 'O': analyserInfo->activity->weekFlag = ODD_WEEKS; return;
		}
	fail:
		yyerror_arg(line, "One of A,E,O letters needed");
	}
	) // WEEKS_ONLY

	AnalFunc(Resource3)	// Resource in Group or RequiredResources.
	{
		NoAttr;
		analyserInfo->resourcesList->push_back(XMLGetText(inDesc, line));
	}

	AnalAttFunc(conjunctive)	// --> analyserInfo->conjunctive
	{
		if (strcmp(attribute.value.c_str(), "yes")==0)
			analyserInfo->conjunctive = 1;
		else
		if (strcmp(attribute.value.c_str(), "no")==0)
			analyserInfo->conjunctive = 0;
		else
			yyerror_arg(attribute.line, "'yes' or 'no' required");
	}

	AnalAtt(GroupType)
		AttOpt(conjunctive)
	EndAnalAtt
	
	AnalAA(Group2)	// Called from analyseGroup which do some more
			// complicated things.
			// --> analyserInfo->resourcesList
		Son2Any(Resource, Resource3),
	EndAnal

	AnalFunc(Group);
	
	Anal(RequiredResources)
		Son2Any(Resource, Resource3),
		SonAny(Group)
	EndAnal
	
	AnalAA(Activity2)	// Called from analyseActivity which do some
				// more complicated things.
				// --> analyserInfo->activity
		Son(Length),
		#if ENABLE_WEEKS
		SonOpt(Week),
		#endif
		SonOpt(TimePreferences),
		Son(RequiredResources),
		SonOptSkip(Name)
	EndAnal

	AnalFunc(Activity);

	Anal(Activities)
		SonAny(Activity)
	EndAnal

	AnalFunc(FirstActivity)
	{
		NoAttr;
		analyserInfo->dependence->firstActStr = XMLGetText(inDesc, line);
	}

	AnalFunc(SecondActivity)
	{
		NoAttr;
		analyserInfo->dependence->secondActStr = XMLGetText(inDesc, line);
	}

	AnalFunc(Operator)
	{
		NoAttr;
		const char *str = XMLGetText(inDesc, line).c_str();
		if (strcmp(str, "Before")==0) analyserInfo->dependence->relation = Dependence::before;
		else if (strcmp(str, "Closely before")==0) analyserInfo->dependence->relation = Dependence::closelyBefore;
		else if (strcmp(str, "After")==0) analyserInfo->dependence->relation = Dependence::after;
		else if (strcmp(str, "Closely after")==0) analyserInfo->dependence->relation = Dependence::closelyAfter;
		else if (strcmp(str, "Concurrently")==0) analyserInfo->dependence->relation = Dependence::concurrently;
		else yyerror_arg(line, "One of dependence operators needed (is \"%s\")", str);
	}
	
	AnalAA(Dependence2)	// Called from analyseDependence which do
				// some more complicated things.
				// --> analyser.dependence
		Son(FirstActivity),
		Son(SecondActivity),
		Son(Operator)
	EndAnal
	
	AnalFunc(Dependence);
	
	Anal(Dependences)
		SonAny(Dependence)
	EndAnal
	
	Anal(Problem)
		Son(General),
		Son(Resources),
		Son(Activities),
		SonOpt(Dependences)
	EndAnal

	Anal(Timetable)
		Son(Problem),
		SonOptSkip(Generator),
		SonOptSkip(Solution)
	EndAnal

	Anal( )
		Son(Timetable)
	EndAnal
	
	AnalInt(StartTime, *analyserInfo->startTime);
		// --> analyserInfo->startTime
	
	Anal(UsedResources)
		Son2Any(Resource, Resource3)
	EndAnal
	
	AnalAA(Activity4)	// Called from analyseActivity3 which do
				// some more complicated things.
		Son(StartTime),
		Son(UsedResources)
	EndAnal

	AnalFunc(Activity3);	// Activity in Solution.
	
	Anal(Solution)
		Son2Any(Activity, Activity3)
	EndAnal

	Anal(Timetable2)	// Timetable which contains Solution.
		SonOptSkip(Problem),
		SonOptSkip(Generator),
		Son(Solution)
	EndAnal

	Anal(2)	// Analyse the solution.
		Son2(Timetable, Timetable2)
	EndAnal
	
	void NormalizeTimePreferences();
	/* Normalyse and check correctness of time preferences for all
	 * activities and resources (for more info see problemspec.h). Set
	 * 'additionalTimeslots' to maximum length of break and correct time
	 * preferences. Additionaly copy 'timePreferences' to
	 * 'initialTimePreferences' and calculate 'mdLength'. */

	void GenerateRequiresAndGroups();
	/* Generate 'requiredResources', 'inActivitiesRe', 'groups',
	 * 'requiredGroups', 'inGroups', 'inActivitesGr',
	 * 'activities.timePreferences' and deletes 'temporaryData'. Earlier
	 * time preferences should be normalized. */

	void NormalizeDependences();
	/* Normalize dependences (reverse after and closelyAfter dependences
	 * and cocurrenly dependences with first activity shorter than
	 * second) and change activities IDs to numbers (in dependences). 
	 * Calculate 'inDependences' for activities. */

	void GenerateGrres();
	/* Calculate 'grresNo', 'contresNo', 'mdresNo', 'grres', 'contres',
	 * 'mdres', 'contCount', 'averageUsage' for all resources. Earlier
	 * 'inGroups' and 'inActivietiesRe' should be calculated. */

	void GenerateGrsel();
	/* Calculate grselNo and grsel for all activities. Earlier
	 * 'requiredGroups' should be calculated. */
	
	void GenerateConflicts();
	/* Calculates 'conflictsWith' in activities. Earlier
	 * 'requiredResources' and 'inActivitiesRe' should be calculated. */

	void FindStringsInSolution();
	/* This method does all the postprocessing while analysing solution.
	 * It converts IDs of activities and resources to its numbers. It
	 * also checks if we have every activity exactly once. */

	void AdditionalTimeslotsInSolution();
	/* Convert times in 'solution' from normal to these with
	 * 'additionalTimeslots' at end of each day. */

	int GetNoById(const map<string,int> &m, const string &str);
	/* Return m[str] or print error when not found. */

public:
	void Initialize(const XML::Content *inDesc);
	/* Takes XML the tree and does all the analysis of input. */

	void AddSolution(const XML::Content *inDesc);
	/* This method analyses the solution XML tree. It saves it in
	 * 'solution'.  You can call the method several times. */
 
};

#undef AnalFuncArgs
#undef GetSons
#undef AnalFunc
#undef Anal
#undef AnalAA
#undef EndAnal
#undef Son2
#undef Son2Opt
#undef Son
#undef SonOpt
#undef SonAny
#undef SonOptSkip
#undef AnalInt
#undef AnalAttFuncArgs 
#undef GetAtts
#undef AnalAttFunc
#undef AnalAtt
#undef EndAnalAtt
#undef Att
#undef AttOpt
#undef NoAttr

} // namespace Analyser

#endif
