/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#include "problemspec.h"
#include "inputparser_my.h"
#include "tools.h"
#include "xxx.h"
#include <map>
#include <cstdio>
#include <cassert>

/* Checking if the solution of the problem is correct and couning it's
 * score. */

/* Trzeba skomentowac CONT */

namespace Analyser{

int CheckTP(int time, const ProblemSpecification::TimePreferences &tp, 
	const ProblemSpecification::Activity &act, int an, int rn)
/* Check if the activity 'act' do not break time preferences 'tp'. 'time' is
 * start time of the activity. Return score comming from breaking the soft
 * preferences. Breaks are not counted into use of resource - breaks can be
 * in forbiden timeslots. */
{
	const int length = act.length;
	int score = 0;
	FOREACH(t, tp.hard)
		if (t>=time && t<time+length)
			if (rn<0)
				yyerror_arg("Hard constraint broken {act %d [%s], time %d}", ACT(an), t);
			else
				yyerror_arg("Hard constraint broken {act %d [%s], res %d [%s], time %d}", ACT(an), RES(rn), t);
	FOR(t, time, time+length-1)
	{
		if (tp.soft[t]==0) continue;
		score += tp.soft[t];
		if (rn<0)
			show(25,("Soft constraint broken {act %d [%s], time %d, value %d}\n", ACT(an), t, tp.soft[t]));
		else
			show(25,("Soft constraint broken {act %d [%s], res %d [%s], time %d, value %d}\n", ACT(an), RES(rn), t, tp.soft[t]));
	}
	WEEKS_ONLY( if (act.weekFlag!=ALL_WEEKS && rn>=0) score/=2; )
	return score;
}

#if ENABLE_CONT
void Resize3Dim(vector<vector<int> > *tab, int s1, int s2, int init)
/* Assuming that 'tab' has 2 empty elements, it is resized to (2)(s1)(s2),
 * and values are set to 'init'. */
{
	REP(a,2)
	{
		tab[a].resize(s1);
		FOREACH(b, tab[a])
			b.resize(s2, init);
	}
}
#endif

int ProblemSpecification::CheckSolution() const 
/* Checking if the solution of the problem is correct and couning it's
 * score. We take solution from 'solution'. 'solution' has each activity
 * exactly once, but (possibly) it is unordered. Unassigned activities has
 * negative 'startTime'. Declared in problemspec.h. */
{
	int score0 = 0,	// unassigned activities (UNASSIGNED_SCORE for each)
	    score1 = 0, // soft time preferences for activities
	    score2 = 0,	// soft time preferences for resources
	    score3 = 0	// broken dependences (soft) (Currently soft
			// dependences are not counted in score from
			// searching, so using soft dependences causes
			// errors. They can be used only for counting
			// result, not for solving.)
	    CONT_ARG(score4 = 0)	// broken 'continuous' (soft)
	    CONT_ARG(score5 = 0),	// broken 'continuous' (hard)
					// (CONTINUOUS_SCORE for each unused
					// slot between activities)
	    score6 = 0; // 'Balanced' constraint
	/* When the score should be counted only in odd or even weeks, then it
	 * is divided by 2. */

	vector<int> start(size(solution));   // start time for each activity
	
	#if ENABLE_CONT
	vector<vector<int> > cont_start[2], cont_stop[2];
	/* Indexed: [week (even/odd)][resource number][day in week]. First
	 * and last timeslot in which the resource is used in a given day. 
	 * (These are global timeslot numbers, not numbers in the day.) When
	 * not used, 'cont_start' is 'noOfSlots' and 'cont_stop' is '0'. */

	vector<int> cont_count[2];	
	/* Indexed: [week (even/odd)][resource number]. Number of timeslots
	 * in which the resource is used. Includes breaks after each
	 * activity. */
	 
	/* Unassigned activities which require the resource are also added
	 * to 'cont_count'. So when we have a hole in a middle of a day and
	 * we have a unassigned activity, while counting the 'continouos'
	 * score it assumed (approximately) that the activity is assigned in
	 * the hole. In such a way, we can get negative score, but negative
	 * scores are treated as zero. */

	Resize3Dim(cont_start, size(resources), daysPerWeek, noOfSlots);
	Resize3Dim(cont_stop, size(resources), daysPerWeek, 0);
	REP(a, 2) cont_count[a].resize(size(resources), 0);
	#endif

	vector<vector<int> > used WEEKS_ONLY([2]);
		// For each resource and timeslot we remember if it is used
		// (including breaks after activity). Value <0 means unused,
		// >=0 means used and is number of the activity.

	vector<vector<int> > dayLoad;
		// For each resource and day we remember in how many
		// timeslots it is used (sum of both weeks, not including
		// breaks).

	WEEKS_ONLY2(
	REP(w, 2) {
		used[w].resize(size(resources));
		FOREACH(a, used[w]) a.resize(noOfSlots, -1);
	},
	used.resize(size(resources));
	FOREACH(a, used) a.resize(noOfSlots, -1);
	)
	dayLoad.resize(size(resources));
	FOREACH(d, dayLoad) d.resize(daysPerWeek);

	FOREACH(sa, solution)	
	{
		CREF(act, activities[sa.actNo]);
		start[sa.actNo] = sa.startTime;
		if (sa.startTime<0)
		{
			/* Unassigned activity. */
			score0 += UNASSIGNED_SCORE;
			show(10,("Unasigned activity %d [%s] !!!!!!!!!!!!!!!!!!!!!!\n", ACT(sa.actNo)));
			
			/* Unassigned activities are also counted in
			 * cont_count (see cont_count definition). */

			#if ENABLE_CONT
			FOREACH(r, act.requiredResources)
			{
				WEEKS_ONLY( if (act.weekFlag&EVEN_WEEKS) )
					cont_count[0][r] += act.length+resources[r].breakLen;
				WEEKS_ONLY( if (act.weekFlag&ODD_WEEKS) )
					cont_count[1][r] += act.length+resources[r].breakLen;
			}
			#endif
			continue;
		}

		/* Add score from activity time preferences. */
		score1 += CheckTP(sa.startTime, act.initialTimePreferences, act, sa.actNo, -1);

		if (sa.startTime/slotsPerDay!=(sa.startTime+act.length-1)/slotsPerDay)
			yyerror_arg("Activity crosses end of day {act %d [%s]}", ACT(sa.actNo));
		
		if (sa.startTime>=noOfSlots)
			yyerror_arg("Wrong timeslot number {act %d [%s]}", ACT(sa.actNo));
		/* Are there too many or too few resources listed? */
		if (size(sa.usedResources) != size(act.requiredResources) + size(act.requiredGroups))
			yyerror_arg("Wrong number of used resources {act %d [%s]}", ACT(sa.actNo));
		
		map<int,int> m; // Map from resource number to number of its
				// group. Resources from requiredResources
				// are mapped to -1. Here are all resources
				// - from requiredResources and from groups
				// in requiredGroups.
		/* Create the map. */
		FOREACH(r, act.requiredResources)
			m.insert(pair<int,int>(r, -1));
		REPV(gi, act.requiredGroups)
		{
			int g = act.requiredGroups[gi];
			FOREACH(r, groups[g].resources)
				m.insert(pair<int,int>(r, gi));
		}
		
		vector<bool> gr_used(size(act.requiredGroups), 0);
			// For each group we remember if there already was a
			// resource from this group.
		
		FOREACH(r, sa.usedResources) {
			
			/* Checking if we do not use resource which we
			 * couldn't. */
			VAR(found, m.find(r));
			if (found==m.end())
				yyerror_arg("Activity uses wrong resource {act %d [%s], res %d [%s]}", ACT(sa.actNo), RES(r));
			
			/* Checking if there are no two resources from the
			 * same group used. */
			int f = found->second;
			if (f>=0)
			{
				if (gr_used[f])
				{
					FOREACH(r2, sa.usedResources)
						if (m.find(r2)->second==f)
							yyerror_arg("Two resources from the same group used {act %d [%s], res %d [%s] and %d [%s]}", ACT(sa.actNo), RES(r2), RES(r));
					assert(false);
				}
				gr_used[f] = 1;
			}
			
			/* Checking if the resource is not occupated by
			 * other activity in the same timeslot. */
			#define lastInDay(a) (((a)/slotsPerDay+1)*slotsPerDay-1)
			FOR(t, sa.startTime, min(sa.startTime+act.length-1+resources[r].breakLen, 
				lastInDay(sa.startTime)))
			{
				WEEKS_ONLY(
				REP(w,2) if (act.weekFlag&(2-w)) { REF(used, used[w]);
				)
				if (used[r][t]>=0)
					yyerror_arg("The same resource used "
					"twice (or no break) {act %d [%s] and"
					" %d [%s], res %d [%s], time %d}", 
					ACT(used[r][t]), ACT(sa.actNo), RES(r), t);
				used[r][t] = sa.actNo;
				WEEKS_ONLY(})
			}
			#undef lastInDay
			dayLoad[r][sa.startTime/slotsPerDay] += act.mdLength;
			
			/* Score from soft preferences for this resource. */
			score2 += CheckTP(sa.startTime, resources[r].timePreferences, act, sa.actNo, r);
			
			/* Calculate 'cont_count', 'cont_start' and
			 * 'cont_stop'. Unassigned activities are added to
			 * 'cont_count' earlier. */
			#if ENABLE_CONT
			WEEKS_ONLY( if (act.weekFlag&EVEN_WEEKS) )
			{
				updateMin(cont_start[0][r][sa.startTime/slotsPerDay], sa.startTime);
				updateMax(cont_stop[0][r][sa.startTime/slotsPerDay], sa.startTime+act.length-1); 
				cont_count[0][r] += act.length;
			}
			WEEKS_ONLY( if (act.weekFlag&ODD_WEEKS) )
			{
				updateMin(cont_start[1][r][sa.startTime/slotsPerDay], sa.startTime);
				updateMax(cont_stop[1][r][sa.startTime/slotsPerDay], sa.startTime+act.length-1); 
				cont_count[1][r] += act.length;
			}
			#endif
		}
	}
	
	/* Check for MaxDays constraint. */
	REPV(r, resources)
		if (resources[r].maxDays<daysPerWeek)
		{
			int days = 0;
			REP(day, daysPerWeek)
			{
				WEEKS_ONLY( REP(w, 2) )
					FOR(t, day*slotsPerDay, (day+1)*slotsPerDay-1)
						if (used WEEKS_ONLY([w])[r][t]>=0)
						{
							++days;
							goto label;
						}
				label:
				continue;
			}
			if (days>resources[r].maxDays)
				yyerror_arg("Resource used in too many days {res %d [%s]}", RES(r));
		}

	/* Check for Balanced constraint. */
	REPV(r, resources)
		if (resources[r].balancedScore>0)
			REP(day, daysPerWeek)
			{
				int sum = max(0, dayLoad[r][day]-resources[r].averageUsage);
				if (sum>0)
					show(25, ("Resource %d [%s] in day %d causes 'balanced' score %d\n", RES(r), day, sum));
				score6 += sum*resources[r].balancedScore;
			}
	
	/* Check if dependences hold. */
	REPV(di, dependences)
	{
		CREF(dep, dependences[di]);
		int s1 = start[dep.firstAct], s2 = start[dep.secondAct];
		if (s1<0 || s2<0) continue;
		int l1 = activities[dep.firstAct].length, l2 = activities[dep.secondAct].length;
		bool bk = false;
		switch (dep.relation)
		{
		case Dependence::before:
			if (s1+l1>s2)
				bk = true;
			break;
		case Dependence::closelyBefore:
			if (s1+l1>s2 || s1/slotsPerDay!=s2/slotsPerDay)
				bk = true;
			break;
		case Dependence::concurrently:
			if (s1>s2 || s1+l1<s2+l2)
				bk = true;
			break;
		default:
			assert(0);
			break;
		}
		if (bk && dep.score<0)
			yyerror_arg("Broken %s dependence %d (%d [%s]: %d-%d, %d [%s]: %d-%d)", dep.relation==Dependence::before?
				"before":dep.relation==Dependence::closelyBefore?"closely before":"concurrently", di, 
				ACT(dep.firstAct), s1, s1+l1-1, ACT(dep.secondAct), s2, s2+l2-1);
		if (bk)
			score3 += dep.score;
	}
	
	/* Calculate score from 'continuouosity'. */
	#if ENABLE_CONT
	REPV(r, resources)
	{
		int cont = resources[r].continuousScore, x = 0;
		if (!cont) continue;
		REP(w, 2)
		{
			int dayCount = 0;
			REP(d, daysPerWeek)
				if (cont_start[w][r][d]<=cont_stop[w][r][d])
				{
					x += cont_stop[w][r][d]-cont_start[w][r][d]+1;
					++dayCount;
				}
			x -= cont_count[w][r]+resources[r].breakLen*
				(size(resources[r].inActivitiesRe)-dayCount);
		}
		/* 'x' is now number of unused timeslots between used. */
		
		x = x*(cont/2);
		if (x>0 && cont>0)
		{
			if (cont==CONTINUOUS_SCORE)
			{
				show(10,("Resource usage is not continuous (res %d [%s], penalty %d) !!!!!!!!!!!!!!!!!!!!!!\n", RES(r), x));
				score5 += x;
			}
			else
				score4 += x;
		}
	}
	#endif
	show(10,("Judgement:\n- Unassigned activities: %d\n- Soft time preferences (activities): %d\n"
		"- Soft time preferences (resources): %d\n- Broken dependences: %d\n"
		CONT_ONLY("- Broken 'continuous' (soft): %d\n- Broken 'continuous' (hard): %d\n"
		"- Violated 'balanced': %d\n"),
		score0, score1, score2, score3 CONT_ARG(score4) CONT_ARG(score5), score6));
	return score0+score1+score2+score3 CONT_ONLY(+score4+score5)+score6;
}

} //namespace Analyser
