%{
/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#include "inputparser_my.h"
#include "xml.h"
#include <cstdlib>
#include <cstdio>
#include <cstdarg>
#include "lexanal.h"
#include "multiproc.h"
#include "xxx.h"

/* Parser for XML files. It is assumed that an element can contain or only
 * text or only subelements. Corresponding header file is inputparser_my.h
 * (inputparser.h is automaticaly created by bison). */

#define YYSTYPE int	// type of additional data returned by lexer (which
			// is number of string in addstring/getstring
#define YYDEBUG 0	// bison internal variable for debuging parser
int yylex();

void xprintf(const char *s, ...)
/* fprintf to stderr2. We need such function in 'show'. It's fefined in
 * defines.h. */
{
	va_list ap;
	va_start(ap, s);
	vfprintf(stderr2, s, ap);
	fflush(stderr2);
	va_end(ap);
}

/* The following two functions shows error message 's' and exits program.
 * 's' should not have trailing '\n', it's added automaticaly. The
 * 'yyerror_arg' function can take additional arguments (like 'printf'). */

void yyerror_arg(const char *s, ...)
{
	va_list ap;
	va_start(ap, s);
	fprintf(stderr2, "**************** ERROR ******************\n");
	if (!openedFile.empty())
		fprintf(stderr2, "%s: ", openedFile.c_str());
	vfprintf(stderr2, s, ap);
	fprintf(stderr2, "\n*****************************************\n");
	fflush(stderr2);
	va_end(ap);
	exit(1);
}

void yyerror_arg(int line, const char *s, ...)
{
	va_list ap;
	va_start(ap, s);
	fprintf(stderr2, "**************** ERROR ******************\n");
	if (!openedFile.empty())
		fprintf(stderr2, "%s: ", openedFile.c_str());
	fprintf(stderr2, "line %d: ", line);
	vfprintf(stderr2, s, ap);
	fprintf(stderr2, "\n*****************************************\n");
	fflush(stderr2);
	va_end(ap);
	exit(1);
}

int yyerror(const char *s)
{
	yyerror_arg(yyline, "%s", s);
	return 0; // Only to remove warning. We exit in yyerror_arg.
}


namespace XML
{
/* Defined later in this file, see there for comments. */
void NewElement(const string &name, int line);
void EnterElement(int line);
void LeaveElement(int line);
void CreateTextContent(const string &name, int line);
void AddAttribute(const string &name, const string &value, int line);
}

//#define __attribute__(x)	// What is it?

%}
%token ID
%token STRING
%token COMMENT
%token NUMBER
%token TEXT
%token OPENENDTAG
%token CLOSESINGLETAG
%%

input:	skip element skip
	;

skip:	/* empty */
	| COMMENT skip
	;

element:
	beginelement '>' { INITIALmode(); } content OPENENDTAG { INSIDEmode(); } ID '>' { INITIALmode(); }
	| beginelement CLOSESINGLETAG { INITIALmode(); }
	;

beginelement:
	'<' { INSIDEmode(); } ID { XML::NewElement(LexYacc::getstring(yylval), yyline); } attributes
	;

attributes:
	/* empty */
	| ID '=' STRING { XML::AddAttribute(LexYacc::getstring(yylval-1), LexYacc::getstring(yylval), yyline); } attributes
	; 

content:
	TEXT { XML::CreateTextContent(LexYacc::getstring(yylval), yyline); }
	| { XML::EnterElement(yyline); } elements { XML::LeaveElement(yyline); }
	;

elements:
	/* empty */
	| elements element
	;

%%

#include <stack>

using namespace std;

int yyline;	// Current line number in lexer.
string openedFile;	// Currently opened file.

/* 'addstring' remembers the arguments and returns its number. Then you can
 * get it back calling 'getstring' with this number. Leading and trailing
 * white spaces are removed. They are used to pass strings between lexer and
 * parser (direct communication is not very easy).  The strings are
 * remembered in 'stringtab0' vector. The vector is created and destroyed in
 * 'parseFile', so you can use this functions only while parsing. */

vector<string> *stringtab0;

int LexYacc::addstring(const string &str)
{
	int a, b;
	for (a=0; a<size(str); ++a)	
		if (str[a]!=' ' && str[a]!='\t' && str[a]!='\n' && str[a]!='\r')
			break;
	for (b=size(str)-1; b>=a; --b)	
		if (str[b]!=' ' && str[b]!='\t' && str[b]!='\n' && str[b]!='\r')
			break;
	stringtab0->push_back(str.substr(a, b-a+1));
//	printf("add %d %s\n",size(*stringtab0),str.c_str());
	return size(*stringtab0);	// 0 is never returned.
}

const string &LexYacc::getstring(int nr)
{
//	extern int yylval;
//	printf("get %d %d %s\n", nr, yylval, (*stringtab0)[nr-1].c_str());
	return (*stringtab0)[nr-1];
}

using namespace XML;

stack<Elements*> elementsStack;
/* Stack with pointers to all open Elements classes. The deepest object is
 * on the top. */

void parseFile(FILE *file, Elements &result)
/* Parse XML file, 'result' should be empty. */
{
	extern FILE* yyin;
	yyin = file;
	elementsStack.push(&result);
	stringtab0 = new vector<string>;
	#if YYDEBUG
	yydebug = 1;
	#endif
	yyline = 1;
	yyparse();
	delete stringtab0;
	elementsStack.pop();
}

namespace XML
{

void NewElement(const string &name, int line)
/* Call it when you see tag opening an element. It creates a element in the
 * current 'Elements' object. */
{
	REF(el, elementsStack.top()->children);
	el.push_back(Element(name, line));
}

/* The functions 'GetCurrentElement', 'EnterElement', 'CreateTextContent',
 * 'AddAttribute' can be called only immediately after 'NewElement'. That
 * means that no calls to 'NewElement', 'EnterElement', 'LeaveElement',
 * 'CreateTextContent' can occur between. Moreover, immediately after
 * creating an element using 'NewElement' you should call one of
 * 'CreateTextContent' or 'EnterElement' functions. 
 * 
 * So after 'NewElement' call several times 'AddAttribute' and then
 * 'EnterElement' or 'CreateTextContent'. */

Element &GetCurrentElement()
/* Returns element created by last 'NewElement'. */
{
	REF(list, elementsStack.top()->children);
	return list[size(list)-1];
}

void EnterElement(int line)
/* Sets the content of the element to 'Elements' and opens this object (next
 * 'NewElement' call will use this object). */
{
	Elements *el = new Elements(line);
	GetCurrentElement().content = el;
	elementsStack.push(el);
}

void CreateTextContent(const string &str, int line)
/* Set content of the last created (by 'NewElement') to TextContent. */
{
	GetCurrentElement().content = new TextContent(str, line);
}

void LeaveElement(int line)
/* Closes 'Elements' object opened by last 'EnterElement'. */
{
	elementsStack.top()->endline = line;
	elementsStack.pop();
}

void AddAttribute(const string &name, const string &value, int line)
/* Add a attribute to the last created (by 'NewElement'). */
{
	REF(att, GetCurrentElement().attributes);
	att.push_back(Attribute(name, value, line));
}

/* The folowing mothods are defined in xml.h. */

/* Classes Connent (TextContent, Elements) and Element have 'Destroy'
 * methods. This methods are something like destructors, they 'Destroy' and
 * then delete subobjects. They should be called before deleting the object. 
 * We use such methods instead of destructors because we want to reuse the
 * same variable, but clear the XML tree which in contains (see main.cc).
 * 
 * IMPORTANT: This classes do not have destructors. You have to mannually
 * call 'Destroy' after using variables of this classes. */

void Elements::Destroy()
{
	FOREACH(a, children)
		a.Destroy();
}

/* 'Content' has methods 'GetText' and 'GetElements' returning the same object
 * in subclass type, when called on wrong object displays error message
 * (occurs when something wrong in readed file). */

const Elements &TextContent::GetElements() const
{
	yyerror_arg(line, "This element should content subelements, but it contents only text");
	return *(Elements*)NULL;
}

const TextContent &Elements::GetText() const
{
	yyerror_arg(line, "This element should content only text, but it contents subelements");
	return *(TextContent*)NULL;
}

} // namespace XML
