%{
/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#include "inputparser.h"
#include "inputparser_my.h"
#include "defines.h"

/* Lexical analyser for XML. Used automaticaly by parser (bison). 
 * 
 * There are two states INITIAL (outside tag) and INSIDE (inside tag). 
 * 
 * It returns one of the following:
 * COMMENT (anything inside <!-- --> or <? >
 * < or = or >
 * OPENENDTAG which is </
 * CLOSESINGLETAG which is />
 * ID - identificator name (inside tag)
 * TEXT - anything outside tag
 * STRING - "quoted" text inside tag
 * 
 * For last tree types 'addstring' is called and yylval contains its result
 * (number of string). */

#define DEBUG0 0	// Debuging system for lexical analyser. (1 or 0)
#define out stderr2

void countNL()
{
	for (char *s = yytext; *s; ++s)
		if (*s=='\n')
			++yyline;
}

using namespace LexYacc;
%}

DIGIT	[0-9]
LETTER  [a-zA-Z_]
ID	{LETTER}({LETTER}|{DIGIT})*
NUMBER	{DIGIT}+

%s INSIDE
%option noyywrap
%option nounput

%%

[ \n\r\t]*		{
			countNL();
			if (DEBUG0) fprintf(out,"0xxx [%s]\n", yytext);
			}

<INITIAL>"<!--"([^-]*|"-"[^-]|"--"[^>])*"-->"	{
			countNL();
			if (DEBUG0) fprintf(out,"1xxx [%s]\n", yytext); 
			return COMMENT;
			}
			
<INITIAL>"<?"[^>]*">"	{
			countNL();
			if (DEBUG0) fprintf(out,"2xxx [%s]\n", yytext); 
			return COMMENT;
			}


<INITIAL>[<] | 
<INSIDE>[=] |
<INSIDE>[>]		{
			if (DEBUG0) fprintf(out,"11xxx [%s]\n", yytext);
			return *yytext;
			}

<INITIAL>"</"		{
			if (DEBUG0) fprintf(out,"4xxx [%s]\n", yytext); 
			return OPENENDTAG;
			}

<INSIDE>"/>"		{
			if (DEBUG0) fprintf(out,"5xxx [%s]\n", yytext); 
			return CLOSESINGLETAG;
			}

<INSIDE>{ID}		{
			if (DEBUG0) fprintf(out,"6xxx [%s]\n", yytext); 
			yylval = addstring(yytext); 
			return ID;
			}

<INITIAL>[^<]*		{
			countNL();
			if (DEBUG0) fprintf(out,"9xxx [%s]\n", yytext); 
			yylval = addstring(yytext); 
			return TEXT;
			}

<INSIDE>[\"][^"]*[\"]	{
			countNL();
			if (DEBUG0) fprintf(out,"12xxx [%s]\n", yytext); 
			{
				int l = strlen(yytext)-1;
				char tmpch = yytext[l];
				yytext[l] = 0;
				yylval = addstring(yytext+1); 
				yytext[l] = tmpch;
			}
			return STRING;
			}

.			{
			yyerror("Illegal character in input");
			}

%%

/* Change state of lexical analyzer. There are two states INITIAL (outside
 * tag) and INSIDE (inside tag). */
void INITIALmode() { BEGIN(INITIAL); }
void INSIDEmode() { BEGIN(INSIDE); }
