/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#include <getopt.h>
#include <cstdio>
#include <cstdlib>
#include "inputparser_my.h"
#include "analyser.h"
#include "search.h"
#include "released.h"	// definition of RELEASED
#include "version.h"	// definition of VERSION
#include "multiproc.h"
#include "solution.h"
#include "config.h"
#include "xxx.h"

FILE *stderr2 = stderr;

int verboseMode = 10;	// How much messages display. More means more. All
			// messages with lower or equal importance level are
			// displayes. 10 is normal, 20 is almost everything.
			// Definition in defines.h

bool showScore = false; // set when -s option is given
bool hasInitialSolution = false;// set when there is solution loaded from file
int initialSolutionScore;	// score of that solution
	// Definitions in solution.h

#define COMPILE_TIME __DATE__ " " __TIME__

char *thisprogramname;

void badopt()
/* Error message when bad command line options. */
{
	fprintf(stderr2, "To get more information about options run `%s --help'.\n", thisprogramname);
	exit(1);
}

FILE *openFile(const char *fname, const char *type)
/* Open file 'fname'. When fname=='-' return stdin / stdout. Display error
 * when impossible. */
{
	if (fname[0]=='-' && !fname[1])
	{
		if (*type=='r')
		{
			#ifdef _WIN32
			stdinAlreadyUsed = true;
			#endif
			openedFile = "stdin";
			return stdin;
		} else {
			openedFile = "stdout";
			return stdout;
		}
	}
	openedFile = string("File '")+fname+"'";
	VAR(file, fopen(fname, type));
	if (file==NULL)
	{
		if (*type=='r')
			fprintf(stderr2, "File '%s' doesn't exist\n", fname);
		else
			fprintf(stderr2, "Couldn't create file '%s'\n", fname);
		badopt();
	}
	return file;
}

void closeFile(FILE *&file)
/* Close file open with openFile. */
{
	if (file!=stdin && file!=stdout)
		fclose(file);
	file = NULL;
	openedFile.clear();
}

Analyser::AnalyserClass ac;
const Analyser::ProblemSpecification &ps = ac;

int main(int argc, char **argv)
{
	char resultFile[1024] = "";	// file to which result is written
	char initFile[1024] = "";	// file with initial solution
	int c;
	bool calculate = false;	// calculate corresponds the '-c' option
	srand(2007);
	
	/* Analyse command line options. Initialize. */
	
	thisprogramname = argv[0];
	struct option options[] = {
		{ "help", no_argument, 0, 'h' },
		{ "version", no_argument, 0, 'x' },
		{ "result", optional_argument, 0, 'r' },
		{ "init", optional_argument, 0, 'i' },
		{ "calculate", no_argument, 0, 'c' },
		{ "verbose", optional_argument, 0, 'v' },
		{ "quiet", no_argument, 0, 'q' },
		{ "score", no_argument, 0, 's' },
		#if ENABLE_MULTI
		{ "processes", required_argument, 0, 'p'},
		#endif
		{ "errors", no_argument, 0, 'e'},
		{ "limit", required_argument, 0, 'l'}
	};
	while ((c = getopt_long(argc, argv, "escqv::r::i::t:l:"MULTI_ONLY("p:"), options, NULL))>=0)
		switch (c) {
		case 'h':
				fprintf(stderr2, ""
"usage: %s [OPTIONS] INPUTFILE OUTPUTFILE\n"
"   Input or output file can be '-' which holds for standard input/output\n"
"      --help                 print this help, then exit\n"
"      --version              print program version number, then exit\n"
"   -l --limit=time           set maximal time of working (in seconds)\n"
MULTI_ONLY("   -p --processes=count      number of paralel processes\n")
"   -v --verbose=level        level of verbosity (normal=10) [=20]\n"
"   -q --quiet                do not show anyting (except errors)\n"
"   -s --score                show score of each better solution to stderr\n"
"   -c --calculate            do not look for solution, only calculate\n"
"                             the score of the existing one\n"
"   -i --init[=file]          start with solution from file [INPUTFILE]\n"
"   -r --result[=file]        write score to file (or to stdout)\n"
"   -e --errors               use stdout instead of stderr\n",
				argv[0]);
				return 0;
		case '?':
			badopt();
		case 'x':
			fprintf(stderr2, "ttsolve-%s (%s)\ncompile time: %s\n", VERSION, RELEASED, COMPILE_TIME);
			return 0;
		case 'e':
			stderr2 = stdout;
			break;
		case 'r':
			strcpy(resultFile, optarg ? optarg : "-");
			break;
		case 'i':
			strcpy(initFile, optarg ? optarg : "");
			hasInitialSolution = true;
			break;
		case 'l':
			sscanf(optarg, "%d", &timeLimit);
			break;
		MULTI_ONLY(
		case 'p':
			sscanf(optarg, "%d", &noOfProcesses);
			break;
		)
		case 'v':
			verboseMode = 20;
			if (optarg) sscanf(optarg, "%d", &verboseMode);
			break;
		case 'q':
			verboseMode = 0;
			break;
		case 'c':
			calculate = true;
			break;
		case 's':
			showScore = true;
			break;
		}
	if (optind+2!=argc) 
		badopt();

	show(10, ("ttsolve-%s (%s, compile time: %s)\n", VERSION, RELEASED, COMPILE_TIME));
	loadConfig(); 
	
	/* Parse and analyse input file. */

	show(10, ("Parsing file '%s'\n", argv[optind]));
	XML::Elements inDesc;
	FILE *f1 = openFile(argv[optind], "r");
	parseFile(f1, inDesc);
	show(10,("Analysing input\n"));
	ac.Initialize(&inDesc);
	closeFile(f1);

	/* Load the initial solution. */

	if (hasInitialSolution)
	{
		if (!*initFile) 
			strcpy(initFile, argv[optind]);
		FILE *f2 = NULL;
		if (strcmp(initFile, argv[optind])!=0)
		{
			show(20,("Destroing XML tree\n"));
			inDesc.Destroy();
			inDesc.children.resize(0);

			show(10, ("Parsing file '%s'\n", initFile));
			FILE *f2 = openFile(initFile, "r");
			parseFile(f2, inDesc);
		}

		show(10,("Analysing solution\n"));
		ac.AddSolution(&inDesc);

		if (f2!=NULL)
			closeFile(f2);
		// it's here because while file is open errors are connected
		// with this file.

		show(10,("Counting result of initial solution\n"));
		initialSolutionScore = ps.CheckSolution();
	}
	
	show(20,("Destroing XML tree\n"));
	inDesc.Destroy();
	inDesc.children.resize(0);

	if (!calculate)
	{
		/* Normal mode. */
	
		performSearch(ps); // Search for solution.

		///* When there are dependences bestSugestion can contain
		// * incorrect solution. */
		//if (ps.dependences.empty())
		
		copySolution(bestSolution, bestSugestion, ps);
		if (showScore)
		{
			fprintf(stderr2, "%d\n", bestScore);
			fflush(stderr2);
		}
	
		/* Save the solution. Solution is also remembered in
		 * ps.solution to be checked*/
	
		show(10,("Saving solution to file '%s'\n", argv[optind+1]));
		VAR(f2, openFile(argv[optind+1], "w"));
		XML::Elements outDesc;
		saveSolution(ps, f2, outDesc);
		closeFile(f2);
		ac.AddSolution(&outDesc);
	}
	else
	{
		/* Only check the existing solution. */
	
		show(10,("Parsing file '%s'\n", argv[optind+1]));
		FILE *f1 = openFile(argv[optind+1], "r");
		parseFile(f1, inDesc);

		show(10,("Analysing solution\n"));
		ac.AddSolution(&inDesc);
		closeFile(f1);

		show(20,("Destroing XML tree\n"));
		inDesc.Destroy();
		inDesc.children.resize(0);
	}

	/* After all check if the result is correct (and recalculate score). */

	show(10,("Counting result\n"));
	int res = ps.CheckSolution();
	show(10,("Result is: %d\n", res));
	if (!calculate && res!=bestScore)
		yyerror_arg("Internal error - score counting");

	/* Save score. */
	if (*resultFile)
	{
		show(10,("Saving result to file '%s'\n", resultFile));
		FILE *file = openFile(resultFile, "w");
		fprintf(file, "%d\n", res);
		closeFile(file);
	}
	#if ENABLE_DEBUG
	extern int prop_work[2], prop_call[2], prop_workmm[2];
	show(10,("Propagators called %d (%d) times, worked %d (%d) times, min/max changed %d (%d) times\n", 
		prop_call[0], prop_call[1], prop_work[0], prop_work[1], prop_workmm[0], prop_workmm[1]));	//--!!
	#endif
}
