/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#include "search.h"
#include "multiproc.h"
#include "solution.h"
#include "problemspec.h"
#include <csignal>
#include "xxx.h"
#ifdef _WIN32
#include <windows.h>
#endif
#if ENABLE_MULTI
#include <unistd.h>
#include <sys/wait.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#endif
#include <climits>

/* Multiprocessing is turned on when ENABLE_MULTI is set (defines.h) and
 * '-p' command line option is given. Number of processes given in '-p'
 * options (saved in noOfProcesses) is only number of the counting
 * processes. Additionaly there is one main process. The counting processes
 * are all childs of the main process. 
 * 
 * Comunication is done using pipes.  There are noOfProcesses+1 pipes, one
 * pipe to every child and one common pipe childs to parent. When a child
 * wants to do something it sends its number. In respond the main program
 * sends the number of activity which should be checked (it is argument of
 * the tryChange function). When -1 is sent to the child process, the
 * process exits.
 * 
 * The first solution is always calculated in the main program. The
 * subprocesses starts after that and they only try to modify it.
 * 
 * Solutions found by subprograms are returned using shared memory. 
 * Variables bestSolution and bestSugestion are in shared memory. When
 * process have something better than we have it writes it to these common
 * variables. When starting a searching operation in a child process
 * bestSugestion i copied into local variable prototypeSolution which will
 * be the template to the process (the bestSugestion can be later changed by
 * other processes, but prototypeSolution can not, so we always use the
 * same). The bestScore is also used to stop searching when we see that we
 * cannot get anything better that the bestSolution.
 * 
 * Access to bestSolution and bestSugestion is protected by semaphores
 * (readers/writers model).
 * 
 * This module allocates memory for bestSolution and bestSugestion even when
 * multiprocessing is not turned on (in local memory in that case). And
 * always allocates prototypeSolution in local memory.
 * 
 * Another functionality of this module is to handle SIGINT and SIGTERM
 * signals (when they apear clearly the program clearly exits writing
 * solution). The handler sets the fastStop variable to 1. Searching
 * functions seeing this are finishing itself and the program end. The
 * signal handler is installed only in the main process. The main process
 * seeing fastStop set, do not send more requests to child processes, but
 * sends -1 (which causes that the subprocesses exits). So the subprocesses
 * exits after they finish current searching. 
 *
 * The signals SIGINT and SIGTERM in child processes are ignored (fastStop
 * is set to 1, but it no not cause anything).
 *
 * Signal handler is installed when calling 'performSearch' (it calls
 * 'initializeMultiproc'). So while analysing input or when in '-c' mode the
 * signals will not be catched.
 * 
 * When you will kill some of the subprocesses the program will be normally
 * working using the others, and at the end the program will clearly exit
 * (but it will try to write to bad pipe !NEED TO BE FIXED!). When you will
 * kill all the subprocesses, it's problem, crash possible. !NEED TO BE
 * FIXED!
 * 
 * Under Windows there are cathed also "ConsoleCtrl" (eg.closing the console
 * window). Additionally, when not reading problem specification from
 * standard input, any characters on standard input closes the program (it's
 * the easiest way to close ttsolve from another program).
 *
 * IMPORTANT: Call InitializeMultiproc even when multiprocessing is turned
 * off. */

int noOfProcesses = 0;	// Number of child processes, 0 when multiprocessing
			// turned off.
int fastStop = 0;	// Set to 1 after SIGTERM or SIGINT signal. 

#if ENABLE_MULTI
vector<int> writePipe, readPipe;
/* noOfProcesses+1 pipes. readPipe has descriptors for reading, writePipe
 * for writing. Child process x can write to pipe noOfProcesses and read
 * from pipe x. Main process can write to pipes form 0 to noOfProcesses-1
 * and read from pipe noOfProcesses. Another will be closed. */
#endif

#ifdef _WIN32
/* Says if stdin is used for reading problem specification or something. If
 * not set, then any characters on stdin will terminate the program (only
 * under Windows). */
bool stdinAlreadyUsed = false;
#endif


MULTI_ONLY(
void subProcess(int processNo, const Analyser::ProblemSpecification &ps, Spaces::MySpace &space)
/* Function executed in child process (should not return). */
{
	/* Close unused pipes. */
	REP(a, noOfProcesses)
		if (a!=processNo)
		{
			close(writePipe[a]);
			close(readPipe[a]);
		}
	close(writePipe[processNo]);
	close(readPipe[noOfProcesses]);
	
	for (;;)
	{
		write(writePipe[noOfProcesses], &processNo, sizeof(int));
		int act;
		if (read(readPipe[processNo], &act, sizeof(int))!=sizeof(int))
			continue;	// read may be interrupted by signal!!!
		show(18,("--- Trying activity: %d [%s]\n", ACT(act)));
		if (act<0)
			exit(0);
		tryChange(ps, space, act);
	}
}
)

void sigKill(int)
/* Handler installed on SIGINT and SIGTERM signals. */
{
	fastStop = 1;
	show(10,("Terminating program, please wait...\n"));
}

#ifdef _WIN32
BOOL CALLBACK sigKill2(DWORD)
/* Handler installed on Windows signals. */
{
	fastStop = 1;
	show(10,("Terminating program, please wait...\n"));
	return TRUE;
}

DWORD WINAPI waitingThread(LPVOID)
{
	getchar();
	raise(SIGTERM);
	return 0;
}
#endif

void assignSolutionPtr(Solution &solution, const Analyser::ProblemSpecification &ps, void *ptr)
/* It assigns memory to 'solution'. 'ptr' is a pointer to memory area of
 * appropiate size (it can be in shared or local memory). Score is set to
 * infinity. */
{
	solution.time = static_cast<int*>(ptr);
	solution.grsel = solution.time+size(ps.activities);
	solution.bestScorePtr = solution.grsel+ps.grselNo;
	*solution.bestScorePtr = INT_MAX;
}

void allocSolution(Solution &solution, const Analyser::ProblemSpecification &ps)
/* Allocate memory for 'solution' in local memory. */
{
	assignSolutionPtr(solution, ps, malloc(sizeof(int)*(size(ps.activities)+ps.grselNo+1)));
}

MULTI_ONLY(
void sharedSolutions(int shmid, const Analyser::ProblemSpecification &ps)
/* Assigns to 'bestSolution' and 'bestSugestion' memory in shared memory area
 * with id 'shmid' (assuming it has appropiate size). */
{
	void *ptr = shmat(shmid, NULL, 0);
	assignSolutionPtr(bestSolution, ps, ptr);
	assignSolutionPtr(bestSugestion, ps, 
		static_cast<void*>(static_cast<int*>(ptr)+size(ps.activities)+ps.grselNo+1));
}

int semid; 
/* Semaprore id. It will be used to locking access to bestSolution and
 * bestSugestion. Initialy set to noOfProcesses. When a process wants to
 * reads decreases it by 1, when wants to write decreases it by
 * noOfProcesses. So when someones wants to write noone can read/write. */

/* Use the following two functions to lock unlock access to bestSolution and
 * bestSugestion. (Reading from bestScore is done without locking, assuming
 * it atomic operation.) */

void lockSolution(int readonly)
{
	if (!noOfProcesses) return;
	sembuf buf;
	buf.sem_num = 0;
	buf.sem_op = readonly ? -1 : -noOfProcesses;
	buf.sem_flg = 0;
	semop(semid, &buf, 1);
}

void unlockSolution(int readonly)
{
	if (!noOfProcesses) return;
	sembuf buf;
	buf.sem_num = 0;
	buf.sem_op = readonly ? 1 : noOfProcesses;
	buf.sem_flg = 0;
	semop(semid, &buf, 1);
}
)


void initializeMultiproc(const Analyser::ProblemSpecification &ps, Spaces::MySpace &space)
{
	// Signal handling.
#ifdef _WIN32
        signal(SIGTERM, sigKill);
        signal(SIGINT, sigKill);
        SetConsoleCtrlHandler(sigKill2, TRUE);
        DWORD threadid;
        CreateThread(NULL, 0, waitingThread, 0, 0, &threadid);
#else
	struct sigaction sact;
	sact.sa_handler = sigKill;
	sigemptyset(&sact.sa_mask);
	sact.sa_flags = 0;
	sigaction(SIGTERM, &sact, NULL);
	sigaction(SIGINT, &sact, NULL);
#endif
                
	allocSolution(prototypeSolution, ps);

	if (!noOfProcesses)
	{
		// When single process mode - allocate memory and return.
		
		allocSolution(bestSolution, ps);
		allocSolution(bestSugestion, ps);
		return;
	}
MULTI_ONLY(
	show(10,("Forking\n"));

	// Allocate shared memory.
	int shmid = shmget(IPC_PRIVATE, 2*sizeof(int)*(size(ps.activities)+ps.grselNo+1), 0600);

	// Create semaphore.
	semid = semget(IPC_PRIVATE, 1, 0600);
	semctl(semid, 0, SETVAL, noOfProcesses);

	// Create pipes.
	readPipe.resize(noOfProcesses+1);
	writePipe.resize(noOfProcesses+1);
	REP(a, noOfProcesses+1)
	{
		int tab[2];
		pipe(tab);
		readPipe[a] = tab[0];
		writePipe[a] = tab[1];
	}

	// Create subprocesses.
	REP(a, noOfProcesses)
		if (!fork()) 
		{
			sharedSolutions(shmid, ps);
			subProcess(a, ps, space);
		}

	// Close unused pipes.
	close(writePipe[noOfProcesses]);
	REP(a, noOfProcesses)
		close(readPipe[a]);

	// Attach shared memory.
	sharedSolutions(shmid, ps);
	
	// Shared memory will be freed when all processes will exit.
	shmctl(shmid, IPC_RMID, NULL);
)
}

MULTI_ONLY(
void tryChangeMultiproc(const Analyser::ProblemSpecification &ps, int act)
/* Called by main process to give a work to one of subprocesses. Call it
 * only when multiprocessing turned on, instead of tryChange. */
{
	int p;
	show(18,("### Waiting for subprocess\n"));
	if (read(readPipe[noOfProcesses], &p, sizeof(int))!=sizeof(int))
		return;		// read may be interrupted by signal!!!
	show(18,("### Trying activity: %d [%s]\n", ACT(act)));
	write(writePipe[p], &act, sizeof(int))!=sizeof(int);
	show(18,("### Write ok\n"));
}
)

void closeMultiproc()
{
	if (!noOfProcesses)
		return;
MULTI_ONLY(
	int m=-1;
	show(18,("### Terminating subprocesses\n"));
	REP(a, noOfProcesses)
	{
		write(writePipe[a], &m, sizeof(int));
		close(writePipe[a]);
	}
	close(readPipe[noOfProcesses]);
	REP(a, noOfProcesses)
		if (wait(NULL)<0)
		{
			--a;
			continue;
		}
	semctl(semid, 0, IPC_RMID, 0);
	show(18,("### Subprocesses finished\n"));
)
}

