/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#ifndef __MYSPACE_H
#define __MYSPACE_H

#include <cassert>
#include "analyser.h"
#include "space.h"

struct Solution;

WEEKS_ONLY(
class int3tab
{
public:
	int tab[3];
	int &operator[](int index) { return tab[index]; }
	int3tab(int x, int y, int z) { tab[0] = x; tab[1] = y; tab[2] = z; }
};
)

namespace Spaces
{

using namespace Analyser;

struct ResScore;

class MySpace : public Space
{
public:
	vector<vector<int> > daysSel;	// [mdres][day] -> number of timeslots allocated (sum of both weeks)
	vector<int> daysNo;	// [mdres] -> number of days used
	vector<int> grsel;	// [grsel] -> resourceNo
	vector<int> waiting;
	int noOfWaiting;
	vector<vector<WEEKS_ONLY2(int3tab, int)> > availableRes; // [group][time]
	vector<vector<char> > isGrresUsed; // [grres][time] -> 0,1,2,3 (when)
	vector<bool> used;	// [time]
	#if ENABLE_CONT
	vector<vector<int> > cont_start WEEKS_ONLY([2]), cont_stop WEEKS_ONLY([2]);	
		// [contres][day], we include breaks after activities
	#endif
	
	static int searchSugested; // do we want to search the sugestedSolution at first (else - normal mode)

private:
DEBUG_ONLY(public:)//!!!!!+++--++

	struct Info
	{
		int act, time;
		stack<vector<int> > res;
	};
	stack<Info> info;
	stack<pair<int, int> > journal_selection;

public:
	int myScore;
	MySpace();
	
	void InitializeTime(int act);

	int PrepareSelection();
	void Select(int pos);
	int PrepareSelection_All();
	int Select_All(int pos);
	void IHaveSolution();
	void AssignActivities(int act, const vector<pair<int,int> > &actList);
	int AreWeSelectingTime() { return info.top().res.empty(); }

	struct JournalState
	{
		int vpos, ppos, cpos;
		int myScore;
		int randseed;
		int noOfWaiting;
		int infosize, inforessize, selectionsize;
		JournalState(int vpos, int ppos, int cpos, int myScore, int randseed, 
		  int noOfWaiting, int infosize, int inforessize, int selectionsize) 
			: vpos(vpos), ppos(ppos), cpos(cpos), myScore(myScore), randseed(randseed),
			  noOfWaiting(noOfWaiting), infosize(infosize), inforessize(inforessize),
			  selectionsize(selectionsize) {}
	};
	
	// Remember current state.
	JournalState GetJournalState()
	{
		WaitStable();
		return JournalState(journal_values.size(), journal_propagators.size(),
			journal_cont.size(), myScore, randseed, noOfWaiting,
			info.size(), info.top().res.size(), journal_selection.size());
	}
	
	// Take the space back to state in js.
	void BackToState(JournalState js DEBUG_ARG(const MySpace *sp = NULL));

private:

	struct ContInfo
	{
		int contres, day;
		WEEKS_ONLY(int week;)
		int start, stop;
		ContInfo(int contres, int day WEEKS_ARG(int week), int start, int stop)
			: contres(contres), day(day) WEEKS_ARG(week(week)), start(start), stop(stop) {}
		ContInfo() {}
	};

	stack<ContInfo> journal_cont; // Stack of changed cont_start, cont_stop boundaries.

	vector<int> time_sel;	// [act] -> time selected (or -1)

	int randseed;
	int Random();
	void ScoreFromSoftPreferences(int from, int to, const PS::TimePreferences &tp WEEKS_ARG(int divide));
	CONT_ONLY( int CalculateContinuousScore(const PS::Resource &res WEEKS_ARG(int which), 
		vector<vector<int> > cont_start WEEKS_ONLY([2]), 
		vector<vector<int> > cont_stop WEEKS_ONLY([2])); )
	CONT_ONLY( void UpdateContinuous(int from, int to, const PS::Resource &res WEEKS_ARG(int which)); )
	void FindAndChooseBestAct();
	void FindBestTime(int act);
	void FindBestResource();
	void FindAllResources();
	void ChooseTime(int t);
	void ChooseResource(int pos);
	void UnuseActivity();
	
	int CalculateBalancedScore(int a, const PS::Resource &res, int day, vector<vector<int> > daysSel)
	/* Return score for violating 'Balanced' constraint for resource
	 * 'res' by activity 'a' when it is allocated in day 'day'. We assume
	 * that 'a' is not included in 'daysSel' yet. */
	{
		assert(res.mdres>=0);
		return max(0, daysSel[res.mdres][day]+ps.activities[a].mdLength-res.averageUsage)
			*res.balancedScore;
	}
	
	void RemoveResource(int res, int from, int to WEEKS_ARG(int flag));
	/* 'r' is number of resource, but it has to be group resource. We
	 * say that this resource is not available from 'from' to 'to' in
	 * odd (flag=1), even (flag=2) or both (flag=3) weeks. Modified
	 * structures: 'isGrresUsed', 'availableRes'. Additionaly activities
	 * using this resource can not use these timeslots. When it is last
	 * resource in group at given time, then activities using the group
	 * can not use these timeslots. */

	void RemoveResourceByAct(int a, int r);
	/* We say that activity 'a' uses resource 'r'. Mark, that the
	 * resource is not available while used by the activity (+break).
	 * Update 'daysSel' and 'daysNo'. Disallow activities dependent from
	 * 'r' to use the timeslots.  Update score from 'Balanced'
	 * constraint. */

	void UndoRemoveResource(int r, int from, int to WEEKS_ARG(int flag));
	/* Undo the 'RemoveResource' operation. 'r' is number of resource,
	 * but it has to be group resource. We say that this resource is
	 * again available from 'from' to 'to' in odd (flag=1), even
	 * (flag=2) or both (flag=3) weeks.  It only modifies structures:
	 * 'isGrresUsed', 'availableRes', but it do not change anything for
	 * activities dependent from this resource. */

	void UndoRemoveResourceByAct(int act, int res);
	/* We say that activity 'act' no more uses resource 'res'. We mark,
	 * that the resource is available again in the time when there was
	 * the activity (+break). It corrects only 'isGrresUsed',
	 * 'availableRes', 'daysSel', 'daysNo' structures, but do not change
	 * anything for activities dependent from this resource. */
	
	void DisjointWithActivities(const vector<int> &actVec, int start, int stop WEEKS_ARG(int flag));
	/* 'vec' is vector of activities numbers. We say these the
	 * activities can not use slots from 'start' to 'stop'. Depending on
	 * 'specialFlag' we take into account only activities using: only
	 * odd (|1), only even (|2), both (|4) weeks. (It is possible, that
	 * we want to remove only resources using both weeks, but leave
	 * resources using only odd weeks and only even weeks
	 * - when different resources are available in odd and even weeks). */

	void ScoreUpdated();
	void FastShufle();
	void GenerateSolution(Solution &bestSolution);

	CONT_ONLY( void FSInsertCont(int from, int to, int r WEEKS_ARG(int which)); )
	CONT_ONLY( void FSRemoveCont(int from, int to, int r WEEKS_ARG(int which)); )
	void FSUpdateResScore(int a, int t0, int len, int dir, ResScore &resScore);
	void FSUpdateMyScoreFromRes(int a, int dir, ResScore &resScore);
	void FSRemoveAct(int a, int oldT, int len, bool breakReq, ResScore &resScore);
	void FSInsertAct(int a, int newT, int len, bool breakReq, ResScore &resScore);
	int FSInsertIfPossible(int a, int newT, int len, bool breakReq, int &ok, ResScore &resScore); // when !ok return first good time
	int FSTryMove(int a, int newT, int &curLength, bool &isBreak, int &ok, ResScore &resScore); // return next t to check
	void FSCreateResScore(int a, int curLength, ResScore &resScore);
	void FSCreateResUsed();
	
	static vector<vector<char> > FSResUsed; //[res][time]
	vector<int> FSgrsel;	// [grsel] -> resourceNo
	#if ENABLE_CONT
	vector<vector<int> > FScont_start WEEKS_ONLY([2]), FScont_stop WEEKS_ONLY([2]);	// [contres][day]
	#endif
	vector<vector<int> > FSdaysSel;	// [mdres][day] -> number of timeslots allocated
	vector<int> FSdaysNo;	// [mdres] -> number of days used

	DEBUG_ONLY(void CheckCurrentSolution();)
};

} // namespace Spaces

#endif
