/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#ifndef __PROBLEMSPEC_H
#define __PROBLEMSPEC_H

#include <vector> 
#include <string>
#include <map>
#include "defines.h"
#include "xxx.h"
using namespace std;

/* Here is class in which we hold the data loaded from input file. There is
 * also field for solution. */

namespace Analyser
{

#if ENABLE_WEEKS
#define ODD_WEEKS 1
#define EVEN_WEEKS 2
#define ALL_WEEKS (ODD_WEEKS|EVEN_WEEKS)
#endif

#define PS ProblemSpecification

class ProblemSpecification
{
	/* IMPORTANT: Some of the variables are not set during the analysis
	 * phase. E.g. resources required by activity are not known (after
	 * the activity is read), because the resources can be not read yet,
	 * so we don't know its numbers. */

public:
	int daysPerWeek, slotsPerDay, noOfSlots;
	int mdresNo;	// Number of resources which have maxDays constraint.
	int grresNo,	// Number of resources which are in at least one group.
	    grselNo	// Number of selections from groups (sum of number
			// of selections from groups for each activity)
	    CONT_ARG(contresNo);	// Number of resources which need to
					// be continuous.

	int additionalTimeslots;
	/* Every day here has 'additioanTimeslots' more than the day in
	 * input file. Using the additional timeslots will be forbiden (hard
	 * preferences for all resources and activities). Motivation:
	 * 'additionalTimeslots' will be equal to maximum length of break. 
	 * Now we do not have to check if the activities are in different
	 * days, always distance between activities have to be at least
	 * length of break. */

	struct TimePreferences
	{
		/* The structure can be in two forms: normalized or not.
		 *
		 * At the beginning it is not normalized. 
		 *
		 * 'soft' contains negative and nonnegative values. It
		 * begins with negative value and after every negative value
		 * there is at least one nonnegative value. Nonnegative
		 * values are number of timeslots, while negative value is
		 * -score for following numbers of timeslots.
		 * 
		 * When 'hardMeansUnavail' is nonzero, then 'hard' is a list
		 * of timeslots which can not be used.  Otherwise 'hard' is
		 * a list of available timeslots.
		 *
		 * 'hardMeansUnavail' is set to 1 when in input file there
		 * are <Hard> elements, or to 0, when there are <Avail>
		 * elements (inside corresponding <TimePreferences>
		 * element). It is set to -1 if we don't know yet or when
		 * there are no hard time preferences for this
		 * activity/resource (it is nonzero, all timeslots are
		 * available).
		 * 
		 * After 'Normalize' is called the structure becomes
		 * normalized. It happens during postprocessing of analysing
		 * (from 'NormalizeTimePreferences'). 'Normalize' also adds
		 * 'addSlots' slots at the end of every day and sets this
		 * slots as forbiden.
		 * 
		 * In normalized form 'soft' has size equal to number of
		 * timeslots, and for each timeslot its value is the score.
		 *
		 * The 'hard' is list of unavailable timeslots, in ascending
		 * order, every exactly once.
		 * 
		 * 'hardMeansUnavail' is set to 1.
		 
		 * After call to 'CompressSoft' (only in activity time
		 * preferences) the compressedSoft for each timeslots gives
		 * score of the whole activity when it is starting at this
		 * timeslot (sum of slots t..t+length-1). */

		vector<int> soft, hard, compressedSoft;
		int hardMeansUnavail;
		TimePreferences() : hardMeansUnavail(-1) {}
		void Normalize(int oldNoOfSlots, int oldSlotsPerDay, int addSlots);
			// Defined in analyser.cc.

		void CompressSoft(int length);
			// Defined in analyser.cc.

		TimePreferences &operator+=(const TimePreferences &tp2);
		/* Sum of TimePreferences means that both hard preferences
		 * must hold and score for soft preferences is sum of score
		 * from each. The operator works correctly in normalized
		 * form. Defined in analyser.cc. */
	};

	struct Resource
	{
		string id;
		TimePreferences timePreferences;
		
		CONT_ONLY( int continuousScore; )
		/* Zero when resource do not need to be continuous. If it
		 * should be continuous, it is equal to score given in input
		 * file (if defined) or to CONTINUOUS_SCORE (otherwise). */
		
		vector<int> inActivitiesRe, inGroups;
		/* List of activities and groups in which it is used. */
		
		int breakLen;	// How many timeslots we need between any
				// two activities using this resource.
		int maxDays;	// Maximum number of days in week in which
				// the resource can be used.
		int mdres;	// Index number in list of resources which
				// have MaxDays or Balanced constraint.
		int grres;	// Index number in list of resources used in
				// groups (or -1 when not in any group).
		int balancedScore;
		int averageUsage;
			// Sum of lengths of all activities using this
			// resource (not in groups) (sum of both weeks)
			// divided by number of allowed days.
		CONT_ONLY(int contres;)	// Index number in list of resources
					// which should be continuous (or -1
					// if it shouldn't).
		CONT_ONLY(int contCount WEEKS_ONLY([2]);)
			// Number of timeslots in which this resource will
			// occur (sum of lengths of activities which require
			// this resource) - separately for both weeks.
			// Includes break after each activity. Group
			// activities can not be continuous, so we know that
			// value exactly.
	};
	vector<Resource> resources;
	
	struct Group
	{
		/* Only disjunctive groups are here. Conjunctive group it is
		 * the same as list of resources - the resources from
		 * conjunctive group are simply added to list of
		 * requiredResources. Here no two groups are identical (it
		 * means list of resource differs). When two groups have the
		 * same resources but in different order we say they are
		 * different. The order of resources is the same as in input
		 * file. We prefer to use earlier resources. */

		vector<int> resources;	// Resources in the group.
		vector<int> activities; // Activities which uses the group.

		int operator<(const Group& gr2) const; 		
		/* Smaller groups are less than biger. Groups of the same
		 * size are in lexicographical order of resources. 
		 * Comparation is used for eliminating identical groups.
		 * Defined in analyser.cc. */
	};
	vector<Group> groups;

	struct Activity
	{
		string id;
		
		TimePreferences timePreferences, initialTimePreferences;
		/* initialTimePreferences are time preferences only for this
		 * activity (as in input file). timePreferences take into
		 * account also time preferences for all resources in
		 * required Resources (but not these in requiredGroups). */
		
		#if ENABLE_WEEKS
		char weekFlag;	// ALL_WEEKS or EVEN_WEEKS or ODD_WEEKS
		#endif
		int length;
		int mdLength;	// 2*length for ALL_WEEKS and non-week
				// version, length for EVEN_WEEKS or
				// ODD_WEEKS.
		int grsel;	// Index number on list of selections from
				// groups. It is index of first group, next
				// groups has next indexes.
		vector<int> requiredResources;
		vector<int> requiredGroups;
		vector<pair<int,int> > conflictsWith;
			// List of activities which uses the same resources
			// in requiredResources. A pair is <activity number,
			// length of break needed>.
		//vector<int> lightConflictsWith;
		vector<int> inDependences;

		struct TemporaryData
		{
			/* During analysis we don't know numbers of
			 * resources or groups yet. So we remember here IDs
			 * of the resources (required directly and in
			 * groups). Later this structure is deleted. */
			 
			vector<string> requiredResources;
			vector<vector<string> > requiredGroups;
		};
		TemporaryData *temporaryData;
		
		void Initialize() { temporaryData = new TemporaryData(); }
		/* 'Initialize' is called while reading the activity. We can
		 * not use constructor, because in vector<Activity> there
		 * probably are some unused objects. */
	};
	vector<Activity> activities;

	struct Dependence
	{
		/* Atter 'NormalizeDependences' is called (it happens during
		 * postprocessing of analysing) there will be no after or
		 * closelyAfter dependences and in concurrently dependences
		 * the first activity will be longer (not shorter) than the
		 * second. */
		
		string firstActStr, secondActStr;
		int firstAct, secondAct;
		/* During analysis we don't know the activity numbers, so we
		 * remenber here their IDs. After 'NormalizeDependenses' the
		 * ...ActStr variables are cleared. */
		
		int score;	// -1 for hard dependences
		enum Relation { before, closelyBefore, after, closelyAfter, concurrently } relation;
			// Closely means in the same day. Concurrently means
			// that shorter (not longer) actitity should start
			// no earlier and finish not later than the second one.
	};
	vector<Dependence> dependences;
	
	struct SolActivity
	{
		/* 'solution' has each activity exactly once, but (possibly)
		 * it is unordered. */
		 
		/* While analysing the solution the resources and activities
		 * IDs are read and later they are converted to resources
		 * and activities numbers and then the usedResourcesStr
		 * vector and id variable are cleared. */

		string id;
		int actNo;
		int startTime; // Unassigned Activities have negative 'startTime'.
		vector<string> usedResourcesStr;
		vector<int> usedResources;
	};
	vector<SolActivity> solution;
	
	int CheckSolution() const;
	/* Checking if the solution of the problem is correct and counting
	 * it's score. We take solution from 'solution'. Defined in
	 * checksol.cc. */


	void SolutionToBestSol() const;	
	/* This function copies 'solution' stored in this object to
	 * 'bestSolution' global variable. (It needs also some coversion.)
	 * Defined in solution.cc.
	 *
	 * IMPORTANT (MULTIPROC): It do not lock 'bestSolution' while
	 * writing to it. Use only while other process are not working (they
	 * are already created, but they have not got any orders yet). */

	/* Shortcuts for getting resource or activity identificator. */
	inline const char *ResId(int r) const { return r>=0&&r<size(resources)?resources[r].id.c_str():"illegal"; }
	inline const char *ActId(int a) const { return a>=0&&a<size(activities)?activities[a].id.c_str():"illegal"; }
	#define RES(r) r, ps.ResId(r)
	#define ACT(a) a, ps.ActId(a)
};

} // namespace Analyser

extern const Analyser::ProblemSpecification &ps;

#endif
