/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#include "propagators.h"
#include "problemspec.h"
#include "space.h"
#include "defines.h"
#include <cstdio>

/* See propagators.h for comments. */

#if ENABLE_DEBUG
int prop_work[2] = {0, 0}, 
    prop_call[2] = {0, 0}, 
    prop_workmm[2] = {0, 0}, 
    prop_change, prop_chmm;	//--!!
#endif

namespace Propagators
{

using namespace Spaces;

#define SHOW_INFO(name)	\
	show(25,("{"#name"Propagator is working (activities: %d [%s] (%d-%d) %s %d [%s] (%d-%d))}\n", \
	ACT(a1), t1.min, t1.max, a1==actNo?"->":"<-", ACT(a2), t2.min, t2.max))

DisjointPropagator::DisjointPropagator(Space &sp, int a10, int l10, int a20, int l20) : a1(a10), a2(a20), l1(l10), l2(l20) 
{ 
	Initialize(sp);
	InitializeVar(sp, a1);
	InitializeVar(sp, a2);
}

bool DisjointPropagator::Propagate(Space &sp, int actNo)
{
	Space::Time &t1 = sp.time[a1], &t2 = sp.time[a2];
	SHOW_INFO(Disjoint);
	DEBUG_ONLY(
	int buf = (t1.min!=t1.max && t2.min!=t2.max);	//--!!
	++prop_call[0];					//--!!
	prop_call[1] += buf;				//--!!
	prop_chmm = prop_change = 0;			//--!!
	)
	if (actNo==a1)
		sp.Exclude(a2, t1.max-l2+1, t1.min+l1-1);
	else
		sp.Exclude(a1, t2.max-l1+1, t2.min+l2-1);
	DEBUG_ONLY(
	if (prop_change)			//--!!
	{					//--!!
		++prop_work[0];			//--!!
		prop_work[1] += buf;		//--!!
	}
	if (prop_chmm)				//--!!
	{					//--!!
		++prop_workmm[0];		//--!!
		prop_workmm[1] += buf;		//--!!
	}
	)
	return (t1.max+l1-1<t2.min || t2.max+l2<t1.min || t1.min==t1.max || t2.min == t2.max);
}

BeforePropagator::BeforePropagator(Space &sp, int a10, int l10, int a20) : a1(a10), a2(a20), l1(l10)
{ 
	Initialize(sp);
	InitializeVar(sp, a1);
	InitializeVar(sp, a2);
}

bool BeforePropagator::Propagate(Space &sp, int actNo)
{
	Space::Time &t1 = sp.time[a1], &t2 = sp.time[a2];
	SHOW_INFO(Before);
	if (actNo==a1)
		sp.Exclude(a2, t2.min, t1.min+l1-1);
	else
		sp.Exclude(a1, t2.max-l1+1, t1.max);
	return (t1.max+l1-1<t2.min || t1.min==t1.max || t2.min == t2.max);
}

SameDayPropagator::SameDayPropagator(Space &sp, int a10, int a20, int slotsPDay) : a1(a10), a2(a20), slotsPerDay(slotsPDay)
{ 
	Initialize(sp);
	InitializeVar(sp, a1);
	InitializeVar(sp, a2);
}

bool SameDayPropagator::Propagate(Space &sp, int actNo)
{
	Space::Time &t1 = sp.time[a1], &t2 = sp.time[a2];
	SHOW_INFO(SameDay);
	if (actNo==a1)
	{
		sp.Exclude(a2, t2.min, t1.min/slotsPerDay*slotsPerDay+1);
		sp.Exclude(a2, t1.max/slotsPerDay*slotsPerDay+slotsPerDay, t2.max);
	}else{
		sp.Exclude(a1, t1.min, t2.min/slotsPerDay*slotsPerDay+1);
		sp.Exclude(a1, t2.max/slotsPerDay*slotsPerDay+slotsPerDay, t1.max);
	}
	return t1.max/slotsPerDay==t1.min/slotsPerDay || t1.max/slotsPerDay==t1.min/slotsPerDay;
}

ConcurrentlyPropagator::ConcurrentlyPropagator(Space &sp, int a10, int a20, int maxD) : a1(a10), a2(a20), maxDiff(maxD)
{ 
	Initialize(sp);
	InitializeVar(sp, a1);
	InitializeVar(sp, a2);
}

bool ConcurrentlyPropagator::Propagate(Space &sp, int actNo)
{
	Space::Time &t1 = sp.time[a1], &t2 = sp.time[a2];
	SHOW_INFO(Concurrently);
	if (actNo==a1)
	{
		sp.Exclude(a2, t2.min, t1.min-1);
		sp.Exclude(a2, t1.max+1+maxDiff, t2.max);
	}else{
		sp.Exclude(a1, t1.min, t2.min-1-maxDiff);
		sp.Exclude(a1, t2.max+1, t1.max);
	}
	return ((t2.max-t1.min<=maxDiff && t1.max<=t2.min) || t1.min==t1.max || t2.min == t2.max);
}

} // namespace Propagators
