/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#include "space.h"
#include "defines.h"
#include "tools.h"
#include "inputparser_my.h"
#include "xxx.h"
#include "problemspec.h"

#if ENABLE_DEBUG
extern int prop_change, prop_chmm;			//--!!
#endif

namespace Spaces
{

using namespace Propagators;

Space::~Space()
{
	REPV(pi, propagators)
		delete propagators[pi];
}

void Space::VariableChanged(int act)
/* Mark that the variable changes (call after direct operations on
 * 'possible'). */ 
{
	if (time[act].modifiedAt<=time[act].actualAt)
	{
		changedVar.push(act);
		show(25,("Variable changed: %d [%s] (%d->%d)\n", ACT(act), time[act].min, time[act].max));
	}
	time[act].modifiedAt = moment++;
}

void Space::WaitStable()
/* Compute all unstable propagators. It should be called before any
 * operations that reads something from the space. */
{
	if (state!=normal) return;
	try
	{
		show(20,("Processing propagators...\n"));
		while (!changedVar.empty())
		{
			int act = changedVar.front();
			changedVar.pop();
			show(25,("{Variable %d [%s] changed}\n", ACT(act)));
			REF(t, time[act]);
			REPV(a, t.propagators)
				if (t.propagators[a]->Propagate(*this, act))
				{
					// remove the propagator
					journal_propagators.push(make_pair(act, t.propagators[a]));
					removeV(t.propagators, a--);
					show(25,("{Removing the propagator}\n"));
				}
			t.actualAt = moment++;
		}
//		FOREACH(gpi, globalprop)			//--++
//			(*gpi)->Propagate(*this, -1);		//--++
	}
	catch (...)
	{
		Fail();
	}
}

void Space::Fail()
/* Call when the space failes (when you've catched the exception). */
{
	state = failed;
}

void Space::UpdateMinMax(int act, bool ch)
/* Recompute the 'min' and 'max' for given activity. When it causes fail,
 * the exception is throwed. Call after changing 'possible'. 'ch' means that
 * the 'min/max' was changed also earlier, so 'VariableChanged' has to be
 * called. */
{
	REF(t, time[act]);
	while (t.max>=0 && !t.possible[t.max]) { t.max--; ch = true; }
	if (t.max<0) 
		throw failno=act;
	while (!t.possible[t.min]) { t.min++; ch = true; }
	if (ch)
	{
		DEBUG_ONLY(prop_chmm = 1;)				//--!!
		VariableChanged(act);
	}
}

bool Space::Exclude(int act, int from, int to)
/* Call when you know that 'act' can not start between 'from' and 'to'. When
 * it causes fail, the exception is throwed. Returns true if 'min' or 'max'
 * changed. */
{
	REF(t, time[act]);
	bool changed = false;
	if (from<t.min) from = t.min;
	if (to>t.max) to = t.max;
	FOR(b, from, to)
		if (t.possible[b])
		{
			changed = true;
			t.possible[b] = false;
			journal_values.push(make_pair(act, b));
		}
	if (!changed) return 0;
	show(25,("{Activity %d [%s] can not start between %d and %d}\n", ACT(act), from, to));
	bool ch2 = false;
	if (t.min>=from && t.min<=to) { t.min = to+1; ch2 = true; }
	if (t.max>=from && t.max<=to) { t.max = from-1; ch2 = true; }
	UpdateMinMax(act, ch2);
	DEBUG_ONLY(prop_change = 1;)				//--!!
	return ch2;
}

void Space::SetVar(int act, int val)
/* Call to set 'act' to specific value 'val'. It is assumed that 'val' is
 * possible value. */
{
	REF(t, time[act]);
	if (!t.possible[val]) 
		yyerror_arg("Internal error");
	if (t.min == t.max && t.min == val) return;
	FOR(a, t.min, t.max)
		if (a!=val && t.possible[a])
		{
			t.possible[a] = false;
			journal_values.push(make_pair(act, a));
		}
	t.min = t.max = val;
	VariableChanged(act);
}

void Space::Time::Initialize(int noOfSlots)
/* Instead of constructor. */
{
	min = 0; 
	max = noOfSlots-1;
	modifiedAt = 1; 
	actualAt = 0;
	possible.resize(noOfSlots, true);
}

Space::Space(int varsNo) : moment(2), state(normal)
{
	time.resize(varsNo);
}

void Space::Succed()
/* Call when you accept current state as solution. */
{
	WaitStable();
	if (state==normal)
		state = succeded;
}


} // namespace Spaces
