/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#ifndef __SPACE_H
#define __SPACE_H

#include "defines.h"
#include <vector>
#include <queue>
#include <stack>

using namespace std;

namespace Propagators { class Propagator; }

namespace Spaces
{

class Space
{
public:
	vector<Propagators::Propagator*> propagators;
		// All propagators.

	/* When a variable's modifiedAt is newer that actualAt, it is in
	 * 'changedVar' queue. */
	queue<int> changedVar;	// Queue with variable numbers.

	//vector<vector<int> > matching;	// For every global AllDisjoint
					// propagator we have here last
					// matching - for each graph node it
					// is number of node connected with
					// it (or -1 if none).	//--++
	
	struct Time
	{
		vector<bool> possible;	// For every timeslot we remember if
					// it is possible or not to start
					// the activite then.
		vector<Propagators::Propagator*> propagators;	
			// Propagators which binds the variable.
		int min, max;	// First and last timeslot for which
				// 'possible' is true.
		int modifiedAt;	// Last moment in which the variable was
				// modified.
		int actualAt;	// Last moment in which the variable's
				// propagators were working.
		void Initialize(int noOfSlots); // Instead of
						// constructor.
	};
	vector<Time> time;	// [activity number]

	enum State { failed, normal, succeded };

	int failno;	// Activity number on witch the fail occured
			// (informative role only)
	int moment;	// Time counter, used in 'modifiedAt' and
			// 'actualAt'. Increased after each use, so every
			// moment is different

	Space(int noOfVars);
	
	bool Exclude(int act, int from, int to);
	/* Call when you know that 'act' can not start between 'from' and
	 * 'to'. When it causes fail, the exception is throwed. Returns true
	 * if 'min' or 'max' changed. */

	void SetVar(int act, int val);
	/* Call to set 'act' to specific value 'val'. It is assumed that
	 * 'val' is possible value. */

	void WaitStable();	
	/* Compute all unstable propagators. It should be called before any
	 * operations that reads something from the space. */
	
	State GetState()
	{
		WaitStable();
		return state;
	}
	
	void VariableChanged(int act);
	/* Mark that the variable changes (call after direct operations on
	 * 'possible'). */

	void UpdateMinMax(int act, bool ch = false);
	/* Recompute the 'min' and 'max' for given activity. When it causes
	 * fail, the exception is throwed. Call after changing 'possible'.
	 * 'ch' means that the 'min/max' was changed also earlier, so
	 * 'VariableChanged' has to be called. */

	~Space();
	
	
protected:
	State state;
	/* State of the space. 'failed' means that for some variables we do
	 * not have possible values. 'succeded' means that 'Succed' was
	 * called - for every variable we have only one value, except some
	 * unassigned variables. On level of this class we can not recognize
	 * which variables are assumed to be unassigned - it is done in
	 * 'MySpace' subslass. Otherwise we have 'normal' state. */

	void Succed();	// Call when you accept current state as solution.
	void Fail();	// Call when the space failes (when you've catched
			// the exception).
	stack<pair<int, int> > journal_values; 
		// Stack of removed (variable, value).
	stack<pair<int, Propagators::Propagator*> > journal_propagators;
		// Stack of removed (variable, propagator).

private:
	Space operator=(const Space &);
		// Not implemented - forbids to use the = operator on
		// spaces (the 'Space' can not be copied).
};

} // namespace Spaces

namespace Propagators
{

/* This is an abstract base class for propagator. A propagator it is
 * something what encures specific relation between activities start times
 * (e.g. that the activities do not overlap) and when something is known
 * about one of them it can deduce something abuot the second.
 * 
 * We say that the propagator is stable when we are sure that no more
 * deductions can be done (calling 'Propagate' do not change anything). 
 * Otherwise we say, that it is unstable.
 *
 * When creating a propagator (from constructor) you should call
 * 'Initialize'. It adds the propagator to the list of propagators in space
 * as unstable propagator. Then you should call 'InitializeVar' for each
 * activity binded by the propagator.
 * 
 * Call 'Propagate' to do deduction connected with the propagators
 * (implemented in subclasses). When it causes fail, the exception will be
 * thrown. */

class Propagator
{
public:
	virtual bool Propagate(Spaces::Space &sp, int actNo) = 0;
	virtual ~Propagator() {}
protected:
	void Initialize(Spaces::Space &sp)
	{
		sp.propagators.push_back(this);
	}

	void InitializeVar(Spaces::Space &sp, int a)
	{
		sp.time[a].propagators.push_back(this);
	}
};

} // namespace Propagators

#endif
