/*
 *  Author: Pawel Parys <parys@mimuw.edu.pl> (2003-2004)
 *
 *  For licence see doc/LICENCE.
 */

#ifndef __XML_H
#define __XML_H

#include <vector>
#include <string>
using namespace std;

namespace XML
{

/* This objects represents XML tree.
 *
 * 'Element' have 'Attribute's and 'Content'. 'Content' can be of two kinds:
 * 'Elements' (which is list of 'Element's) or 'TextContent'.
 * 
 * Methods 'Show' are used internally by xmlgenerator.cc.
 * 
 * 'Content' has methods 'GetText' and 'GetElements' returning the same
 * object in subclass type, when called on wrong object displays error
 * message (occurs when something wrong in readed file).
 * 
 * There are also 'Destroy' methods. This methods are something like
 * destructors, they 'Destroy' and then delete subobjects. They should be
 * called before deleting the object. We use such methods instead of
 * destructors because we want to reuse the same variable, but clear the XML
 * tree which in contains (see main.cc).
 *
 * IMPORTANT: This classes do not have destructors. You have to mannually
 * call 'Destroy' after using variables of this classes.
 * 
 * Implementation of this this header is in file xmlgenerator.cc ('Show'
 * methods) and in file inputparser.y ('Destroy', 'GetText', 'GetElements').
 */

class Attribute
{
public:
	int line;
	string name;
	string value;
	Attribute(const string &n, const string &v, int line = 0) : line(line), name(n), value(v) {}
	void Show() const;
};

class Elements;
class TextContent;

class Content
{
public:
	virtual void Destroy() = 0;
	virtual const Elements &GetElements() const = 0;
	virtual const TextContent &GetText() const = 0;
	virtual void Show() const = 0;
};

class Element
{
public:
	int line;
	string name;
	Content *content;
	vector<Attribute> attributes;
	void Destroy() { if (content!=NULL) { content->Destroy(); delete content; } }
	Element() : content(NULL) {}
	Element(const string &n, int line = 0) : line(line), name(n), content(NULL) {}
	void Show() const;
};

class TextContent : public Content
{
public:
	int line;
	string value;
	TextContent(const string &str, int line = 0) : line(line), value(str) {}
	virtual ~TextContent() {}
	virtual void Destroy() {};
	virtual const Elements &GetElements() const;
	virtual const TextContent &GetText() const { return *this; }
	virtual void Show() const;
};

class Elements : public Content
{
public:
	int line, endline;
	vector<Element> children;
	Elements(int line = 0) : line(line) {}
	virtual ~Elements() {}
	virtual void Destroy();
	virtual const Elements &GetElements() const { return *this; }
	virtual const TextContent &GetText() const;
	virtual void Show() const;
};

} // namespace XML

#endif
